# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import os
import numpy as np

from hyperspy.io import load

my_path = os.path.dirname(__file__)


class TestStackBuilder:

    def test_load_stackbuilder_imagestack(self):
        image_stack = load(os.path.join(my_path, "dm_stackbuilder_plugin",
                                        "test_stackbuilder_imagestack.dm3"))
        data_dimensions = image_stack.data.ndim
        am = image_stack.axes_manager
        axes_dimensions = am.signal_dimension + am.navigation_dimension
        assert data_dimensions == axes_dimensions
        md = image_stack.metadata
        assert md.Acquisition_instrument.TEM.acquisition_mode == "STEM"
        np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_current, 0.0)
        np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
        np.testing.assert_allclose(md.Acquisition_instrument.TEM.camera_length, 15.0)
        np.testing.assert_allclose(
            md.Acquisition_instrument.TEM.dwell_time, 0.03000005078125)
        np.testing.assert_allclose(md.Acquisition_instrument.TEM.magnification, 200000.0)
        assert md.Acquisition_instrument.TEM.microscope == "JEM-ARM200F"
        assert md.General.date == "2015-05-17"
        assert md.General.original_filename == "test_stackbuilder_imagestack.dm3"
        assert md.General.title == "stackbuilder_test4_16x2"
        assert md.General.time == "17:00:16"
        assert md.Sample.description == "DWNC"
        assert md.Signal.quantity == "Electrons (Counts)"
        assert md.Signal.signal_type == ""
        assert md.Signal.binned == False
        np.testing.assert_allclose(md.Signal.Noise_properties.Variance_linear_model.gain_factor,
                        0.15674974)
        np.testing.assert_allclose(md.Signal.Noise_properties.Variance_linear_model.gain_offset,
                        2228741.5)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

"""Interactive widgets that can be added to `Signal` plots.

Example
-------

"""


from hyperspy.drawing.widget import (
    WidgetBase, DraggableWidgetBase, ResizableDraggableWidgetBase,
    Widget2DBase, Widget1DBase, ResizersMixin)
from hyperspy.drawing._widgets.horizontal_line import HorizontalLineWidget
from hyperspy.drawing._widgets.vertical_line import VerticalLineWidget
from hyperspy.drawing._widgets.label import LabelWidget
from hyperspy.drawing._widgets.scalebar import ScaleBar
from hyperspy.drawing._widgets.circle import CircleWidget
from hyperspy.drawing._widgets.rectangles import RectangleWidget, SquareWidget
from hyperspy.drawing._widgets.range import ModifiableSpanSelector, RangeWidget
from hyperspy.drawing._widgets.line2d import Line2DWidget



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

# Set the PyQt API to 2 to avoid incompatibilities between matplotlib
# traitsui

import logging
_logger = logging.getLogger(__name__)
from hyperspy.logger import set_log_level
from hyperspy.defaults_parser import preferences
set_log_level(preferences.General.logging_level)
from hyperspy import datasets
from hyperspy.utils import *
from hyperspy.io import load
from hyperspy import signals
from hyperspy.Release import version as __version__
from hyperspy import docstrings

__doc__ = """

All public packages, functions and classes are available in this module.

%s

Functions:

    create_model
        Create a model for curve fitting.

    get_configuration_directory_path
        Return the configuration directory path.

    load
        Load data into BaseSignal instances from supported files.

    preferences
        Preferences class instance to configure the default value of different
        parameters. It has a CLI and a GUI that can be started by execting its
        `gui` method i.e. `preferences.gui()`.

    stack
        Stack several signals.

    interactive
        Define operations that are automatically recomputed on event changes.

    set_log_level
        Convenience function to set HyperSpy's the log level.


The :mod:`~hyperspy.api` package contains the following submodules/packages:

    :mod:`~hyperspy.api.signals`
        `Signal` classes which are the core of HyperSpy. Use this modules to
        create `Signal` instances manually from numpy arrays. Note that to
        load data from supported file formats is more convenient to use the
        `load` function.
    :mod:`~hyperspy.api.model`
        Contains the :mod:`~hyperspy.api.model.components` module with
        components that can be used to create a model for curve fitting.
    :mod:`~hyperspy.api.eds`
        Functions for energy dispersive X-rays data analysis.
    :mod:`~hyperspy.api.material`
        Useful functions for materials properties and elements database that
        includes physical properties and X-rays and EELS energies.
    :mod:`~hyperspy.api.plot`
        Plotting functions that operate on multiple signals.
    :mod:`~hyperspy.api.datasets`
        Example datasets.
    :mod:`~hyperspy.api.roi`
        Region of interests (ROIs) that operate on `BaseSignal` instances and
        include widgets for interactive operation.
    :mod:`~hyperspy.api.samfire`
        SAMFire utilities (strategies, Pool, fit convergence tests)


For more details see their doctrings.

""" % docstrings.START_HSPY

# Remove the module to avoid polluting the namespace
del docstrings





def get_configuration_directory_path():
    import hyperspy.misc.config_dir
    return hyperspy.misc.config_dir.config_path



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np
from hyperspy._components.expression import Expression


sigma2fwhm = 2 * np.sqrt(2 * np.log(2))


class Gaussian2D(Expression):
    r"""Normalized 2D elliptical Gaussian function component.

    .. math::

        f(x,y) = \frac{A}{2\pi s_x s_y}\exp\left[-\frac{\left(x-x_0\right)
        ^{2}}{2s_{x}^{2}}\frac{\left(y-y_0\right)^{2}}{2s_{y}^{2}}\right]


    =============== =============
    Variable         Parameter
    =============== =============
    :math:`A`        A
    :math:`s_x,s_y`  sigma_x/y
    :math:`x_0,y_0`  centre_x/y
    =============== =============


    Parameters
    ----------
    A : float
        Amplitude (height of the peak scaled by :math:`2 \pi s_x s_y`).
    sigma_x : float
        Width (scale parameter) of the Gaussian distribution in `x` direction.
    sigma_y : float
        Width (scale parameter) of the Gaussian distribution in `y` direction.
    centre_x : float
        Location of the Gaussian maximum (peak position) in `x` direction.
    centre_x : float
        Location of the Gaussian maximum (peak position) in `y` direction.
    add_rotation : bool
        If True, add the parameter `rotation_angle` corresponding to the angle
        between the `x` and the horizontal axis.

    Attributes
    ----------
    fwhm_x, fwhm_x : float
        Convenience attributes to get and set the full-with-half-maxima along
        the two axes.

    Properties
    ----------
    sigma_major : float
        The sigma value of the major axis (axis with the largest sigma value).
    sigma_minor : float
        The sigma value of the minor axis (axis with the smallest sigma value).
    ellipticity : float
        Ratio between the major and minor axis.
    rotation_major_axis : float
        Rotation angle in radian between the major axis (axis with the largest
        sigma value) and the horizontal axis.
        Only for Gaussian2D component created with `add_rotation=True`.
    rotation_angle_wrapped : float
        Rotation angle in radian wrapped to [0, 2*pi].
        Only for Gaussian2D component created with `add_rotation=True`.
    """

    def __init__(self, A=1., sigma_x=1., sigma_y=1., centre_x=0.,
                 centre_y=0, module="numexpr", **kwargs):
        super(Gaussian2D, self).__init__(
            expression="A * (1 / (sigma_x * sigma_y * 2 * pi)) * \
                       exp(-((x - centre_x) ** 2 / (2 * sigma_x ** 2) \
                       + (y - centre_y) ** 2 / (2 * sigma_y ** 2)))",
            name="Gaussian2D",
            A=A,
            sigma_x=sigma_x,
            sigma_y=sigma_y,
            centre_x=centre_x,
            centre_y=centre_y,
            position=("centre_x", "centre_y"),
            module=module,
            autodoc=False,
            **kwargs)

        # Boundaries
        self.A.bmin = 0.

        self.sigma_x.bmin = 0.
        self.sigma_y.bmin = 0.

        self.isbackground = False
        self.convolved = True

    @property
    def fwhm_x(self):
        return self.sigma_x.value * sigma2fwhm

    @fwhm_x.setter
    def fwhm_x(self, value):
        self.sigma_x.value = value / sigma2fwhm

    @property
    def fwhm_y(self):
        return self.sigma_y.value * sigma2fwhm

    @fwhm_y.setter
    def fwhm_y(self, value):
        self.sigma_y.value = value / sigma2fwhm

    @property
    def sigma_major(self):
        if self.sigma_x.value >= self.sigma_y.value:
            return self.sigma_x.value
        else:
            return self.sigma_y.value

    @property
    def sigma_minor(self):
        if self.sigma_x.value >= self.sigma_y.value:
            return self.sigma_y.value
        else:
            return self.sigma_x.value

    @property
    def ellipticity(self):
        return self.sigma_major / self.sigma_minor

    @property
    def rotation_angle_wrapped(self):
        """ Angle in radian wrapped from 0 to 2*pi. """
        return math.fmod(self.rotation_angle.value, 2 * np.pi)

    @property
    def rotation_major_axis(self):
        """Angle between major axis and x-axis, in radian."""
        if self.sigma_x.value >= self.sigma_y.value:
            return self.rotation_angle_wrapped
        else:
            return self.rotation_angle_wrapped - np.pi / 2



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest

from hyperspy._signals.signal1d import Signal1D
from hyperspy.component import Component, Parameter
from hyperspy.components1d import Gaussian, Lorentzian, ScalableFixedPattern


def remove_empty_numpy_strings(dic):
    for k, v in dic.items():
        if isinstance(v, dict):
            remove_empty_numpy_strings(v)
        elif isinstance(v, list):
            for vv in v:
                if isinstance(vv, dict):
                    remove_empty_numpy_strings(vv)
                elif isinstance(vv, np.string_) and len(vv) == 0:
                    vv = ''
        elif isinstance(v, np.string_) and len(v) == 0:
            del dic[k]
            dic[k] = ''


class DummyAxesManager:
    navigation_shape = [1, ]
    navigation_size = 2
    indices = ()

    @property
    def _navigation_shape_in_array(self):
        return self.navigation_shape[::-1]


class TestParameterDictionary:

    def setup_method(self, method):
        self.par = Parameter()
        self.par.name = 'asd'
        self.par._id_name = 'newone'
        self.par.twin_function_expr = "x * x"
        self.par.twin_inverse_function_expr = "sqrt(x)"
        self.par._axes_manager = DummyAxesManager()
        self.par._create_array()
        self.par.value = 1
        self.par.std = 0.1
        self.par.store_current_value_in_array()
        self.par.ext_bounded = False
        self.par.ext_force_positive = False

    def test_to_dictionary(self):
        d = self.par.as_dictionary()

        assert d['name'] == self.par.name
        assert d['_id_name'] == self.par._id_name
        np.testing.assert_equal(d['map']['values'][0], 1)
        np.testing.assert_equal(d['map']['std'][0], 0.1)
        assert d['map']['is_set'][0]
        np.testing.assert_equal(d['value'], self.par.value)
        np.testing.assert_equal(d['std'], self.par.std)
        assert d['free'] is self.par.free
        assert d['self'] == id(self.par)
        np.testing.assert_equal(d['_bounds'], self.par._bounds)
        assert d['ext_bounded'] is self.par.ext_bounded
        assert (
            d['ext_force_positive'] is self.par.ext_force_positive)

    def test_load_dictionary(self):
        d = self.par.as_dictionary()
        p = Parameter()
        p._id_name = 'newone'
        _id = p._load_dictionary(d)

        assert _id == id(self.par)
        assert p.name == self.par.name
        assert p._id_name == self.par._id_name
        np.testing.assert_equal(p.map['values'][0], 1)
        np.testing.assert_equal(p.map['std'][0], 0.1)
        assert p.map['is_set'][0]
        np.testing.assert_equal(p.value, self.par.value)
        np.testing.assert_equal(p.std, self.par.std)
        np.testing.assert_equal(p.free, self.par.free)
        np.testing.assert_equal(p._bounds, self.par._bounds)

        rn = np.random.random()
        np.testing.assert_equal(
            p.twin_function(rn),
            self.par.twin_function(rn))
        np.testing.assert_equal(
            p.twin_inverse_function(rn),
            self.par.twin_inverse_function(rn))

    def test_invalid_name(self):
        d = self.par.as_dictionary()
        d['_id_name'] = 'otherone'
        p = Parameter()
        p._id_name = 'newone'
        with pytest.raises(ValueError):
            _ = p._load_dictionary(d)


class TestComponentDictionary:

    def setup_method(self, method):
        self.parameter_names = ['par1', 'par2']
        self.comp = Component(self.parameter_names)
        self.comp.name = 'newname!'
        self.comp._id_name = 'dummy names yay!'
        self.comp._axes_manager = DummyAxesManager()
        self.comp._create_arrays()
        self.comp.par1.value = 2.
        self.comp.par2.value = 5.
        self.comp.par1.std = 0.2
        self.comp.par2.std = 0.5
        self.comp.store_current_parameters_in_map()

    def test_to_dictionary(self):
        d = self.comp.as_dictionary()
        c = self.comp

        assert c.name == d['name']
        assert c._id_name == d['_id_name']
        assert not d['active_is_multidimensional']
        assert d['active']
        assert d['_active_array'] is None
        for ip, p in enumerate(c.parameters):
            assert p.as_dictionary() == d['parameters'][ip]

        c.active_is_multidimensional = True
        d1 = c.as_dictionary()
        assert d1['active_is_multidimensional']
        np.testing.assert_array_equal(d1['_active_array'], c._active_array)

    def test_load_dictionary(self):
        c = self.comp
        c.par1.twin_function_expr = "x + 2"
        c.par2.twin_function_expr = "x - 2"
        d = c.as_dictionary(True)
        n = Component(self.parameter_names)

        n._id_name = 'dummy names yay!'
        _ = n._load_dictionary(d)
        assert c.name == n.name
        assert c.active == n.active
        assert (
            c.active_is_multidimensional ==
            n.active_is_multidimensional)

        for pn, pc in zip(n.parameters, c.parameters):
            rn = np.random.random()
            assert pn.twin_function(rn) == pc.twin_function(rn)
            assert (
                pn.twin_inverse_function(rn) ==
                pc.twin_inverse_function(rn))
            dn = pn.as_dictionary()
            del dn['self']
            dc = pc.as_dictionary()
            del dc['self']
            print(list(dn.keys()))
            print(list(dc.keys()))
            assert dn == dc

    def test_invalid_component_name(self):
        c = self.comp
        d = c.as_dictionary()
        n = Component(self.parameter_names)
        with pytest.raises(ValueError):
            _ = n._load_dictionary(d)

    def test_invalid_parameter_name(self):
        c = self.comp
        d = c.as_dictionary()
        n = Component([a + 's' for a in self.parameter_names])
        n._id_name = 'dummy names yay!'
        with pytest.raises(ValueError):
            _ = n._load_dictionary(d)


class TestModelDictionary:

    def setup_method(self, method):
        s = Signal1D(np.array([1.0, 2, 4, 7, 12, 7, 4, 2, 1]))
        m = s.create_model()
        m.low_loss = (s + 3.0).deepcopy()
        self.model = m
        self.s = s

        m.append(Gaussian())
        m.append(Gaussian())
        m.append(ScalableFixedPattern(s * 0.3))
        m[0].A.twin = m[1].A
        m.fit()

    def test_to_dictionary(self):
        m = self.model
        d = m.as_dictionary()

        print(d['low_loss'])
        np.testing.assert_allclose(m.low_loss.data, d['low_loss']['data'])
        np.testing.assert_allclose(m.chisq.data, d['chisq.data'])
        np.testing.assert_allclose(m.dof.data, d['dof.data'])
        np.testing.assert_equal(
            d['free_parameters_boundaries'],
            m.free_parameters_boundaries)
        assert d['convolved'] is m.convolved

        for num, c in enumerate(m):
            tmp = c.as_dictionary()
            remove_empty_numpy_strings(tmp)
            assert d['components'][num]['name'] == tmp['name']
            assert d['components'][num]['_id_name'] == tmp['_id_name']
        np.testing.assert_equal(d['components'][-1]['signal1D'],
                                (m.signal * 0.3)._to_dictionary())

    def test_load_dictionary(self):
        d = self.model.as_dictionary()
        mn = self.s.create_model()
        mn.append(Lorentzian())
        mn._load_dictionary(d)
        mo = self.model

        # assert_true(np.allclose(mo.signal1D.data, mn.signal1D.data))
        np.testing.assert_allclose(mo.chisq.data, mn.chisq.data)
        np.testing.assert_allclose(mo.dof.data, mn.dof.data)

        np.testing.assert_allclose(mn.low_loss.data, mo.low_loss.data)

        np.testing.assert_equal(
            mn.free_parameters_boundaries,
            mo.free_parameters_boundaries)
        assert mn.convolved is mo.convolved
        for i in range(len(mn)):
            assert mn[i]._id_name == mo[i]._id_name
            for po, pn in zip(mo[i].parameters, mn[i].parameters):
                np.testing.assert_allclose(po.map['values'], pn.map['values'])
                np.testing.assert_allclose(po.map['is_set'], pn.map['is_set'])

        assert mn[0].A.twin is mn[1].A



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import gc
from os import remove
from unittest import mock

import numpy as np
import pytest

from hyperspy._signals.signal1d import Signal1D
from hyperspy.components1d import Expression, Gaussian, GaussianHF
from hyperspy.io import load


def clean_model_dictionary(d):
    for c in d['components']:
        for p in c['parameters']:
            del p['self']
    return d


class TestModelStoring:

    def setup_method(self, method):
        s = Signal1D(range(100))
        m = s.create_model()
        m.append(Gaussian())
        m.fit()
        self.m = m

    def test_models_getattr(self):
        m = self.m
        s = m.signal
        m.store()
        assert s.models.a is s.models['a']

    def test_models_stub_methods(self):
        m = self.m
        s = m.signal
        m.store()
        s.models.pop = mock.MagicMock()
        s.models.remove = mock.MagicMock()
        s.models.restore = mock.MagicMock()
        s.models.a.restore()
        s.models.a.remove()
        s.models.a.pop()

        assert s.models.pop.call_count == 1
        assert s.models.remove.call_count == 1
        assert s.models.restore.call_count == 1

        assert s.models.pop.call_args[0] == ('a',)
        assert s.models.remove.call_args[0] == ('a',)
        assert s.models.restore.call_args[0] == ('a',)

    def test_models_pop(self):
        m = self.m
        s = m.signal
        m.store()
        s.models.remove = mock.MagicMock()
        s.models.restore = mock.MagicMock()
        s.models.pop('a')
        assert s.models.remove.call_count == 1
        assert s.models.restore.call_count == 1
        assert s.models.remove.call_args[0] == ('a',)
        assert s.models.restore.call_args[0] == ('a',)

    def test_model_store(self):
        m = self.m
        m.store()
        d = m.as_dictionary(True)
        np.testing.assert_equal(
            d,
            m.signal.models._models.a._dict.as_dictionary())

    def test_actually_stored(self):
        m = self.m
        m.store()
        m[0].A.map['values'][0] += 13.33
        m1 = m.signal.models.a.restore()
        assert m[0].A.map['values'] != m1[0].A.map['values']

    def test_models_restore_remove(self):
        m = self.m
        s = m.signal
        m.store('a')
        m1 = s.models.restore('a')
        m2 = s.models.a.restore()
        d_o = clean_model_dictionary(m.as_dictionary())
        d_1 = clean_model_dictionary(m1.as_dictionary())
        d_2 = clean_model_dictionary(m2.as_dictionary())
        np.testing.assert_equal(d_o, d_1)
        np.testing.assert_equal(d_o, d_2)
        assert 1 == len(s.models)
        s.models.a.remove()
        assert 0 == len(s.models)

    def test_store_name_error1(self):
        s = self.m.signal
        with pytest.raises(KeyError):
            s.models.restore('a')

    def test_store_name_error2(self):
        s = self.m.signal
        with pytest.raises(KeyError):
            s.models.restore(3)

    def test_store_name_error3(self):
        s = self.m.signal
        with pytest.raises(KeyError):
            s.models.restore('_a')

    def test_store_name_error4(self):
        s = self.m.signal
        with pytest.raises(KeyError):
            s.models.restore('a._dict')

    def test_store_name_error5(self):
        s = self.m.signal
        self.m.store('b')
        with pytest.raises(KeyError):
            s.models.restore('a')


class TestModelSaving:

    def setup_method(self, method):
        s = Signal1D(range(100))
        m = s.create_model()
        m.append(Gaussian(A=13))
        m[-1].name = 'something'
        m.append(GaussianHF(module="numpy"))
        m[-1].height.value = 3
        m.append(Expression(name="Line", expression="a * x + b", a=1, c=0, rename_pars={"b": "c"}))
        self.m = m

    def test_save_and_load_model(self):
        m = self.m
        m.save('tmp.hdf5', overwrite=True)
        s = load('tmp.hdf5')
        assert hasattr(s.models, 'a')
        mr = s.models.restore('a')
        assert mr.components.something.A.value == 13
        assert mr.components.GaussianHF.height.value == 3
        assert mr.components.Line.a.value == 1
        assert mr.components.Line.c.value == 0
        assert mr.components.Line.function(10) == 10

    def teardown_method(self, method):
        gc.collect()        # Make sure any memmaps are closed first!
        remove('tmp.hdf5')


class TestEELSModelSaving:

    def setup_method(self, method):
        s = Signal1D(range(100))
        s.axes_manager[0].offset = 280
        s.set_signal_type("EELS")
        s.add_elements(["C"])
        s.set_microscope_parameters(100, 10, 10)
        m = s.create_model(auto_background=False)
        m.components.C_K.fine_structure_smoothing = 0.5
        m.components.C_K.fine_structure_width = 50
        m.components.C_K.fine_structure_active = True
        self.m = m

    def test_save_and_load_model(self):
        m = self.m
        m.save('tmp.hdf5', overwrite=True)
        l = load('tmp.hdf5')
        assert hasattr(l.models, 'a')
        n = l.models.restore('a')
        assert n[0].fine_structure_width == 50

    def teardown_method(self, method):
        gc.collect()        # Make sure any memmaps are closed first!
        remove('tmp.hdf5')



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import logging

from hyperspy.drawing.widgets import ResizableDraggableWidgetBase
from hyperspy.drawing.utils import picker_kwargs


_logger = logging.getLogger(__name__)


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)


def angle_between(v1, v2):
    """ Returns the angle in radians between the vectors 'v1' and 'v2'.

    Examples
    --------
    >>> angle_between((1, 0), (0, 1))
    1.5707963267948966
    >>> angle_between((1, 0), (1, 0))
    0.0
    >>> angle_between((1, 0), (-1, 0))
    3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    angle = np.arctan2(v2_u[1], v2_u[0]) - np.arctan2(v1_u[1], v1_u[0])
    # angle = np.arccos(np.dot(v1_u, v2_u))
    if np.isnan(angle):
        if (v1_u == v2_u).all():
            return 0.0
        else:
            return np.pi
    return angle


class Line2DWidget(ResizableDraggableWidgetBase):

    """A free-form line on a 2D plot. Enables dragging and moving the end
    points, but also allows rotation of the widget by moving the mouse beyond
    the end points of the line.

    The widget adds the 'linewidth' attribute, which is different from the size
    in the following regards: 'linewidth' is simply the width of the patch
    drawn from point to point. If 'size' is greater than 1, it will in
    principle select a rotated rectangle. If 'size' is greater than 4, the
    bounds of this rectangle will be visualized by two dashed lines along the
    outline of this rectangle, instead of a single line in the center.

    The widget also adds the attributes 'radius_resize', 'radius_move' and
    'radius_rotate' (defaults: 5, 5, 10), which determines the picker radius
    for resizing, aka. moving the edge points (by picking within
    'radius_resize' from an edge point); for moving (by picking within
    'radius_move' from the body of the line); and for rotation (by picking
    within 'radius_rotate' of the edge points on the "outside" of the line).
    The priority is in the order resize, rotate, move; so the 'radius_rotate'
    should always be larger than 'radius_resize' if the function is to be
    accessible (putting it lower is an easy way to disable the functionality).


    Notes
    -----
    This widget's internal position does not lock to axes points by default.

    Notes
    -----
    The 'position' is now a 2D tuple: tuple(tuple(x1, x2), tuple(y1, y2))
    
    Notes
    -----
    The 'size' property corresponds to line width, so it has a len() of only
    one.
    """

    # Bitfield values for different mouse interaction functions
    FUNC_NONE = 0       # Do nothing
    FUNC_MOVE = 1       # Move the widget
    FUNC_RESIZE = 2     # Move a vertex
    FUNC_ROTATE = 4     # Rotate
    FUNC_SIZERS = 8     # Change linewidth by indicators
    FUNC_A = 32         # Resize/rotate by first vertex
    FUNC_B = 64         # Resize/rotate by second vertex

    def __init__(self, axes_manager, **kwargs):
        super(Line2DWidget, self).__init__(axes_manager, **kwargs)
        self.linewidth = 1
        self.radius_move = self.radius_resize = 5
        self.radius_rotate = 15
        self._mfunc = self.FUNC_NONE    # Mouse interaction function
        self._prev_pos = None
        self._orig_pos = None
        self.snap_all = False
        self._width_indicator_patches = []
        self._size = np.array([0])
        # Set default axes
        if self.axes_manager is not None:
            if self.axes_manager.navigation_dimension > 1:
                self.axes = self.axes_manager.navigation_axes[0:2]
            else:
                self.axes = self.axes_manager.signal_axes[0:2]
        value = self.axes[0].scale if self.axes_manager else 1
        # [[x0, y0], [x1, y1]]
        self._pos = np.array([[0, 0], [value, 0]])

    def _set_size(self, value):
        """Setter for the 'size' property.

        Calls _size_changed to handle size change, if the value has changed.

        """
        value = value[0]  # in this method, value is a float/int
        if value < 0:
            value = 0
        elif value:
            # The size must not be smaller than the scale
            value = np.maximum(value, self.axes[0].scale)
            if self.snap_size:
                value = self._do_snap_size(value)[0]
        if self._size[0] != value:
            if not value and self._size:
                self._size = np.array((0,))
                self._remove_size_patch()
            elif value and not self._size:
                self._size = np.array((value,))
                self._set_size_patch()
            else:
                self._size = np.array((value,))
            self._size_changed()

    def _set_axes(self, axes):
        # _set_axes overwrites self._size so we back it up
        size = self._size
        position = self._pos
        super(Line2DWidget, self)._set_axes(axes)
        # Restore self._size
        self._size = size
        self._pos = position

    def connect_navigate(self):
        raise NotImplementedError("2D lines cannot be used to navigate (yet?)")

    def _validate_pos(self, pos):
        """Make sure all vertices are within axis bounds.
        """
        if np.shape(pos)[1] != len(self.axes):
            raise ValueError()

        ndim = np.shape(pos)[1]
        if ndim != len(self.axes):
            raise ValueError()
        pos = np.maximum(pos, [ax.low_value for ax in self.axes])
        pos = np.minimum(pos, [ax.high_value for ax in self.axes])
        if self.snap_position:
            pos = self._do_snap_position(pos)
        return pos

    def _do_snap_position(self, value=None):
        value = np.array(value) if value is not None else self._pos

        ret1 = super(Line2DWidget, self)._do_snap_position(value[0, :])
        ret2 = super(Line2DWidget, self)._do_snap_position(value[1, :])

        return np.array([ret1, ret2])

    def _set_snap_size(self, value):
        if value and self.axes[0].scale != self.axes[1].scale:
            _logger.warning('Snapping the width of the line is not supported '
                            'for axes with different scale.')
            return
        super()._set_snap_size(value)

    def _do_snap_size(self, value=None):
        if value is None:
            value = self._size[0]
        if hasattr(value, '__len__'):
            value = value[0]
        ax = self.axes[0]  # take one axis, different axis scale not supported
        value = round(value / ax.scale) * ax.scale

        # must return an array to be consistent with the widget API
        return np.array([value])

    def _get_line_normal(self):
        v = np.diff(self._pos, axis=0)   # Line vector
        x = -v[:, 1]
        y = v[:, 0]
        n = np.array([x, y]).T                    # Normal vector
        return n / np.linalg.norm(n)            # Normalized

    def get_line_length(self):
        """Returns line length in axes coordinates. Requires units on all axes
        to be the same to make any physical sense.
        """
        return np.linalg.norm(np.diff(self._pos, axis=0), axis=1)

    def get_centre(self):
        """Get the line center, which is simply the mean position of its
        vertices.
        """
        return np.mean(self._pos, axis=0)

    def _get_width_indicator_coords(self):
        """
        Get coordinates of width indicators.

        The returned format is:
            [[[x0A, y0A], [x1A, y1A]], [[x0B, y0B], [x1B, y1B]]]
            Where A and B refer to the two lines
        """
        n = self.size[0] * self._get_line_normal() / 2.
        c = np.array(self._pos)
        return c + n, c - n

    def _update_patch_position(self):
        self._update_patch_geometry()

    def _update_patch_size(self):
        self._update_patch_geometry()

    def _update_patch_geometry(self):
        """Set line position, and set width indicator's if appropriate
        """
        if self.is_on() and self.patch:
            self.patch[0].set_data(np.array(self._pos).T)
            # Update width indicator if present
            if self._width_indicator_patches:
                wc = self._get_width_indicator_coords()
                for i in range(2):
                    self.patch[1 + i].set_data(wc[i].T)
            self.draw_patch()

    def _set_patch(self):
        """Creates the line, and also creates the width indicators if
        appropriate.
        """
        self.ax.autoscale(False)   # Prevent plotting from rescaling
        xy = np.array(self._pos)
        max_r = max(self.radius_move, self.radius_resize,
                    self.radius_rotate)
        kwargs = picker_kwargs(max_r)
        self.patch = self.ax.plot(
            xy[:, 0], xy[:, 1],
            linestyle='-',
            animated=self.blit,
            lw=self.linewidth,
            c=self.color,
            alpha=self.alpha,
            marker='s',
            markersize=self.radius_resize,
            mew=0.1,
            mfc='lime',
            **kwargs,)[0:1]

    def _set_size_patch(self):
        wc = self._get_width_indicator_coords()
        kwargs = picker_kwargs(self.radius_move)
        for i in range(2):
            wi, = self.ax.plot(
                wc[i][0], wc[i][1],
                linestyle=':',
                animated=self.blit,
                lw=self.linewidth,
                c=self.color,
                **kwargs)
            self.patch.append(wi)
            self._width_indicator_patches.append(wi)

    def _remove_size_patch(self):
        if not self._width_indicator_patches:
            return
        for patch in self._width_indicator_patches:
            self.patch.remove(patch)
            patch.remove()
        self._width_indicator_patches = []

    def _get_vertex(self, event):
        """Check bitfield on self.func, and return vertex index.
        """
        if self.func & self.FUNC_A:
            return 0
        elif self.func & self.FUNC_B:
            return 1
        else:
            return None

    def _get_func_from_pos(self, cx, cy):
        """Get interaction function from pixel position (cx,cy)
        """
        if not self.patch:
            return self.FUNC_NONE

        trans = self.ax.transData
        p = np.array(trans.transform(self._pos))

        # Calculate the distances to the vertecies, and find nearest one
        r2 = np.sum(np.power(p - np.array((cx, cy)), 2), axis=1)
        mini = np.argmin(r2)    # Index of nearest vertex
        minr2 = r2[mini]        # Distance squared to nearest vertex
        del r2
        # Check for resize: Click within radius_resize from edge points
        radius = self.radius_resize
        if minr2 <= radius ** 2:
            ret = self.FUNC_RESIZE
            ret |= self.FUNC_A if mini == 0 else self.FUNC_B
            return ret

        # Check for rotate: Click within radius_rotate on outside of edgepts
        radius = self.radius_rotate
        A = p[0, :]  # Vertex A
        B = p[1, :]  # Vertex B. Assumes one line segment only.
        c = np.array((cx, cy))   # mouse click position
        t = np.dot(c - A, B - A)    # t[0]: A->click, t[1]: A->B
        bas = np.linalg.norm(B - A)**2
        if minr2 <= radius**2:   # If within rotate radius
            if t < 0.0 and mini == 0:   # "Before" A on the line
                return self.FUNC_ROTATE | self.FUNC_A
            elif t > bas and mini == 1:  # "After" B on the line
                return self.FUNC_ROTATE | self.FUNC_B

        # Check for move: Click within radius_move from any point on the line
        radius = self.radius_move
        if 0 < t < bas:
            # A + (t/bas)*(B-A) is closest point on line
            if np.linalg.norm(A + (t / bas) * (B - A) - c) < radius:
                return self.FUNC_MOVE

        # Check for line width resize: Click within radius_move from width
        # indicator lines
        radius = self.radius_move
        wc = self._get_width_indicator_coords()
        for i in range(2):
            A = np.array(trans.transform(wc[i][0]))
            B = np.array(trans.transform(wc[i][1]))
            t = np.dot(c - A, B - A)
            bas = np.linalg.norm(B - A)**2
            if 0 < t < bas:
                # A + (t/bas)*(B-A) is closest point on line
                if np.linalg.norm(A + (t / bas) * (B - A) - c) < radius:
                    return self.FUNC_SIZERS
        return self.FUNC_NONE

    def onpick(self, event):
        """Pick, and if picked, figure out which function to apply. Also store
        pouse position for use by _onmousemove. As rotation does not work very
        well with incremental rotations, the original points are stored if
        we're rotating.
        """
        super(Line2DWidget, self).onpick(event)
        if self.picked:
            me = event.mouseevent
            self.func = self._get_func_from_pos(me.x, me.y)
            self._drag_start = [me.xdata, me.ydata]

    def _onmousemove(self, event):
        """Delegate to _move(), _resize() or _rotate().
        """
        if self.picked is True:
            if self.func & self.FUNC_MOVE and event.inaxes:
                self._move(event)
            elif self.func & self.FUNC_RESIZE and event.inaxes:
                self._resize(event)
            elif self.func & self.FUNC_ROTATE:
                self._rotate(event)
            elif self.func & self.FUNC_SIZERS and event.inaxes:
                self._width_resize(event)

    def _get_diff(self, event):
        """Get difference in position in event and what is stored in _prev_pos,
        in value space.
        """
        if event.xdata is None:
            dx = 0
        else:
            dx = event.xdata - self._drag_start[0]
        if event.ydata is None:
            dy = 0
        else:
            dy = event.ydata - self._drag_start[1]
        return np.array((dx, dy))

    def _move(self, event):
        """Move line by drag start position + difference in mouse post from
        when dragging started (picked).
        """
        dx = self._get_diff(event)
        self.position = self._drag_store[0] + dx

    def _resize(self, event):
        """Move vertex by difference from pick / last mouse move. Update
        '_prev_pos'.
        """
        ip = self._get_vertex(event)
        dx = self._get_diff(event)
        p = np.array(self._pos)     # Copy
        p[ip, 0:2] = self._drag_store[0][ip] + dx
        self.position = p

    def _rotate(self, event):
        """Rotate original points by the angle between mouse position and
        rotation start position (rotation center = line center).
        """
        if None in (event.xdata, event.ydata):
            return
        # Get difference in mouse pos since drag start (picked)
        dx = self._get_diff(event)

        # Rotation should happen in screen position, as anything else will
        # mix units
        trans = self.ax.transData
        scr_zero = np.array(trans.transform((0, 0)))
        dx = np.array(trans.transform(dx)) - scr_zero

        # Get center point = center of original line
        c = trans.transform(np.mean(self._drag_store[0], axis=0))

        # Figure out theta
        v1 = (event.x, event.y) - c     # Center to mouse
        v2 = v1 - dx                    # Center to start pos
        theta = angle_between(v2, v1)   # Rotation between start and mouse

        if event.key is not None and 'shift' in event.key:
            base = 30 * np.pi / 180
            theta = base * round(float(theta) / base)

        # vector from points to center
        w1 = c - trans.transform(self._drag_store[0])
        # rotate into w2 for next point
        w2 = np.array((w1[:, 0] * np.cos(theta) - w1[:, 1] * np.sin(theta),
                       w1[:, 1] * np.cos(theta) + w1[:, 0] * np.sin(theta)))
        self.position = trans.inverted().transform(c + np.rot90(w2))

    def _width_resize(self, event):
        if None in (event.xdata, event.ydata) or self.size[0] == 0:
            return
        # Get difference in mouse pos since drag start (picked)
        dx = self._get_diff(event)
        # Project onto normal axis (dot product onto normal)
        n = self._get_line_normal()
        dn = 2 * np.dot(n, dx)
        if self._selected_artist is self.patch[2]:
            dn *= -1
        self.size = np.abs(self._drag_store[1] + dn)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import matplotlib.pyplot as plt

from hyperspy.drawing.widgets import Widget2DBase, ResizersMixin


class CircleWidget(Widget2DBase, ResizersMixin):

    """CircleWidget is a symmetric, Cicle-patch based widget, which can
    be dragged, and resized by keystrokes/code.
    """

    def __init__(self, axes_manager, **kwargs):
        super(CircleWidget, self).__init__(axes_manager, **kwargs)
        self.size_step = 1.0
        self.size_snap_offset = (0.5 + 1e-8)

    def _set_axes(self, axes):
        super(CircleWidget, self)._set_axes(axes)
        if self.axes:
            self._size[0] = (0.5 + 1e-8) * self.axes[0].scale
            if len(self.axes) > 1:
                self._size[1] = 0

    def _do_snap_size(self, value=None):
        # Snap to odd diameters = ?.5 radius
        value = np.array(value) if value is not None else self._size
        snap_offset = self.size_snap_offset * self.axes[0].scale
        snap_spacing = self.axes[0].scale * self.size_step
        for i in range(2):
            value[i] = max(0, (round((value[i] - snap_offset) / snap_spacing) *
                               snap_spacing + snap_offset))
        return value

    def _set_size(self, value):
        """Setter for the 'size' property. Calls _size_changed to handle size
        change, if the value has changed.
        """
        # Override so that r_inner can be 0
        value = np.minimum(value,
                           [0.5 * ax.size * ax.scale for ax in self.axes])
        # Changed from base:
        min_sizes = np.array(((0.5 + 1e-8) * self.axes[0].scale, 0))
        value = np.maximum(value, min_sizes)
        if value[0] < value[1]:
            self._set_size(value[::-1])
        else:
            if self.snap_size:
                value = self._do_snap_size(value)
            if np.any(self._size != value):
                self._size = value
                self._size_changed()

    def increase_size(self):
        """Increment all sizes by one step. Applied via 'size' property.
        """
        s = np.array(self.size)
        if self.size[1] > 0:
            s += self.size_step * self.axes[0].scale
        else:
            s[0] += self.size_step * self.axes[0].scale
        self.size = s

    def decrease_size(self):
        """Decrement all sizes by one step. Applied via 'size' property.
        """
        s = np.array(self.size)
        if self.size[1] > 0:
            s -= self.size_step * self.axes[0].scale
        else:
            s[0] -= self.size_step * self.axes[0].scale
        self.size = s

    def get_centre(self):
        return self.position

    def _get_patch_xy(self):
        """Returns the xy coordinates of the patch. In this implementation, the
        patch is centered on the position.
        """
        return self.position

    def _set_patch(self):
        """Sets the patch to a matplotlib Circle with the correct geometry.
        The geometry is defined by _get_patch_xy, and size.
        """
        super(CircleWidget, self)._set_patch()
        xy = self._get_patch_xy()
        ro, ri = self.size
        self.patch = [plt.Circle(
            xy, radius=ro,
            animated=self.blit,
            fill=False,
            lw=self.border_thickness,
            ec=self.color,
            alpha=self.alpha,
            picker=True,)]
        if ri > 0:
            self.patch.append(
                plt.Circle(
                    xy, radius=ro,
                    animated=self.blit,
                    fill=False,
                    lw=self.border_thickness,
                    ec=self.color,
                    alpha=self.alpha,
                    picker=True,))

    def _validate_pos(self, value):
        """Constrict the position within bounds.
        """
        value = (min(value[0], self.axes[0].high_value - self._size[0] +
                     (0.5 + 1e-8) * self.axes[0].scale),
                 min(value[1], self.axes[1].high_value - self._size[0] +
                     (0.5 + 1e-8) * self.axes[1].scale))
        value = (max(value[0], self.axes[0].low_value + self._size[0] -
                     (0.5 + 1e-8) * self.axes[0].scale),
                 max(value[1], self.axes[1].low_value + self._size[0] -
                     (0.5 + 1e-8) * self.axes[1].scale))
        return super(CircleWidget, self)._validate_pos(value)

    def get_size_in_indices(self):
        return np.array(self._size / self.axes[0].scale)

    def _update_patch_position(self):
        if self.is_on() and self.patch:
            self.patch[0].center = self._get_patch_xy()
            if self.size[1] > 0:
                self.patch[1].center = self.patch[0].center
            self._update_resizers()
            self.draw_patch()

    def _update_patch_size(self):
        if self.is_on() and self.patch:
            ro, ri = self.size
            self.patch[0].radius = ro
            if ri > 0:
                # Add the inner circle
                if len(self.patch) == 1:
                    # Need to remove the previous patch before using
                    # `_add_patch_to`
                    self.ax.artists.remove(self.patch[0])
                    self.patch = []
                    self._add_patch_to(self.ax)
                self.patch[1].radius = ri
            self._update_resizers()
            self.draw_patch()

    def _update_patch_geometry(self):
        if self.is_on() and self.patch:
            ro, ri = self.size
            self.patch[0].center = self._get_patch_xy()
            self.patch[0].radius = ro
            if ri > 0:
                self.patch[1].center = self.patch[0].center
                self.patch[1].radius = ri
            self._update_resizers()
            self.draw_patch()

    def _onmousemove(self, event):
        'on mouse motion move the patch if picked'
        if self.picked is True and event.inaxes:
            x = event.xdata
            y = event.ydata
            if self.resizer_picked is False:
                x -= self.pick_offset[0]
                y -= self.pick_offset[1]
                self.position = (x, y)
            else:
                rad_vect = np.array((x, y)) - self._pos
                radius = np.sqrt(np.sum(rad_vect**2))
                s = list(self.size)
                if self.resizer_picked < 4:
                    s[0] = radius
                else:
                    s[1] = radius
                self.size = s

    def _get_resizer_pos(self):
        positions = []
        indices = (0, 1) if self.size[1] > 0 else (0, )
        for i in indices:
            r = self._size[i]
            rsize = self._get_resizer_size() / 2
            rp = np.array(self._get_patch_xy())
            p = rp - (r, 0) - rsize             # Left
            positions.append(p)
            p = rp - (0, r) - rsize             # Top
            positions.append(p)
            p = rp + (r, 0) - rsize             # Right
            positions.append(p)
            p = rp + (0, r) - rsize             # Bottom
            positions.append(p)
        return positions



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.learn.svd_pca import svd_solve


def whiten_data(X, centre=True, method="PCA", epsilon=1e-10):
    """Centre and whiten the data X.

    A whitening transformation is used to decorrelate
    the variables, such that the new covariance matrix
    of the whitened data is the identity matrix.

    If X is a random vector with non-singular covariance
    matrix C, and W is a whitening matrix satisfying
    W^T W = C^-1, then the transformation Y = W X will
    yield a whitened random vector Y with unit diagonal
    covariance. In ZCA whitening, the matrix W = C^-1/2,
    while in PCA whitening, the matrix W is the
    eigensystem of C. More details can be found in [Kessy2015]_.

    Parameters
    ----------
    X : numpy array, shape (m, n)
        The input data.
    centre : bool, default True
        If True, centre the data along the features axis.
        If False, do not centre the data.
    method : {"PCA", "ZCA"}
        How to whiten the data. The default is PCA whitening.
    epsilon : float
        Small floating-point value to avoid divide-by-zero errors.

    Returns
    -------
    Y : numpy array, shape (m, n)
        The centred and whitened data.
    W : numpy array, shape (n, n)
        The whitening matrix.

    References
    ----------
    .. [Kessy2015] A. Kessy, A. Lewin, and K. Strimmer, "Optimal
        Whitening and Decorrelation", arXiv:1512.00809, (2015),
        https://arxiv.org/pdf/1512.00809.pdf

    """
    Y = X

    # Centre the variables
    if centre:
        Y -= Y.mean(axis=0)

    # Calculate the whitening matrix
    R = (Y.T @ Y) / Y.shape[0]
    U, S, _ = svd_solve(R, svd_solver="full")
    S = np.sqrt(S + epsilon)[:, np.newaxis]

    if method == "PCA":
        # PCA whitening was the default in HyperSpy < 1.6.0,
        # we keep it as the default here.
        W = U.T / S

    elif method == "ZCA":
        W = U @ (U.T / S)

    else:
        raise ValueError(f"method must be one of ['PCA', 'zca'], got {method}")

    # Whiten the data
    Y = Y @ W.T

    return Y, W



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.components1d import DoublePowerLaw


def test_function():
    g = DoublePowerLaw()
    g.A.value = 3
    g.r.value = 2
    g.origin.value = 1
    g.shift.value = 2
    g.ratio.value = 2
    assert np.isinf(g.function(1))
    assert np.isinf(g.function(3))
    assert g.function(-1) == 0
    assert g.function(0) == 0
    assert g.function(2) == 9
    np.testing.assert_allclose(g.function(10), 0.15948602)
    assert g.grad_A(2) == 3
    np.testing.assert_allclose(g.grad_r(4), -0.3662041)
    assert g.grad_origin(2)  == -6
    assert g.grad_shift(2)  == -12
    assert g.grad_ratio(2)  == 3



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy.  If not, see <http://www.gnu.org/licenses/>.
import hyperspy.roi
from hyperspy.roi import (CircleROI, Line2DROI, Point1DROI, Point2DROI,
                          RectangularROI, SpanROI)

__doc__ = hyperspy.roi.__doc__



import numpy as np
import pytest

from hyperspy.learn.ornmf import ornmf
from hyperspy.signals import Signal1D


def compare_norms(a, b, tol=5e-3):
    assert a.shape == b.shape

    m, n = a.shape
    tol *= m * n
    n1 = np.linalg.norm(a)
    n2 = np.linalg.norm(b)

    assert np.linalg.norm((a / n1) - (b / n2)) < tol


class TestRNMF:
    def setup_method(self, method):
        # Define shape etc.
        m = 128  # Dimensionality
        n = 256  # Number of samples
        r = 3
        s = 0.01

        rng = np.random.RandomState(101)
        U = rng.uniform(0, 1, (m, r))
        V = rng.uniform(0, 1, (n, r))
        X = U @ V.T
        np.divide(X, max(1.0, np.linalg.norm(X)), out=X)

        E = 100 * rng.binomial(1, s, X.shape)
        Y = X + E

        self.m = m
        self.n = n
        self.rank = r
        self.U = U
        self.V = V
        self.X = X
        self.Y = Y
        self.E = E

    @pytest.mark.parametrize("project", [True, False])
    def test_default(self, project):
        W, H = ornmf(self.X, self.rank, project=project)
        compare_norms(W @ H, self.X)

        assert W.shape == self.U.shape
        assert H.shape == self.V.T.shape

    def test_batch_size(self):
        W, H = ornmf(self.X, self.rank, batch_size=2)
        compare_norms(W @ H, self.X)

        assert W.shape == self.U.shape
        assert H.shape == self.V.T.shape

    def test_store_error(self):
        Xhat, Ehat, W, H = ornmf(self.X, self.rank, store_error=True)
        compare_norms(Xhat, self.X)

        assert Xhat.shape == self.X.shape
        assert Ehat.shape == self.E.shape

    def test_corrupted_default(self):
        W, H = ornmf(self.Y, self.rank)
        compare_norms(W @ H, self.X)

    def test_robust(self):
        W, H = ornmf(self.X, self.rank, method="RobustPGD")
        compare_norms(W @ H, self.X)

    def test_corrupted_robust(self):
        W, H = ornmf(self.Y, self.rank, method="RobustPGD")
        compare_norms(W @ H, self.X)

    def test_no_method(self):
        with pytest.raises(ValueError, match=f"'method' not recognised"):
            _ = ornmf(self.X, self.rank, method="uniform")

    def test_subspace_tracking(self):
        W, H = ornmf(self.X, self.rank, method="MomentumSGD")
        compare_norms(W @ H, self.X)

    @pytest.mark.parametrize("subspace_learning_rate", [1.0, 1.1])
    def test_subspace_tracking_learning_rate(self, subspace_learning_rate):
        W, H = ornmf(
            self.X,
            self.rank,
            method="MomentumSGD",
            subspace_learning_rate=subspace_learning_rate,
        )
        compare_norms(W @ H, self.X)

    @pytest.mark.parametrize("subspace_momentum", [0.5, 0.9])
    def test_subspace_tracking_momentum(self, subspace_momentum):
        W, H = ornmf(
            self.X, self.rank, method="MomentumSGD", subspace_momentum=subspace_momentum
        )
        compare_norms(W @ H, self.X)

        with pytest.raises(ValueError, match=f"must be a float between 0 and 1"):
            _ = ornmf(self.X, self.rank, method="MomentumSGD", subspace_momentum=1.9)

    @pytest.mark.parametrize("poisson", [True, False])
    def test_signal(self, poisson):
        # Note that s1.decomposition() operates on the transpose
        # i.e. (n_samples, n_features).
        x = self.Y.T.copy().reshape(16, 16, 128)

        if poisson:
            x -= x.min()
            x[x <= 0] = 1e-16

        s1 = Signal1D(x)

        X_out, E_out = s1.decomposition(
            normalize_poissonian_noise=poisson,
            algorithm="ORNMF",
            output_dimension=self.rank,
            return_info=True,
        )

        # Check the low-rank component MSE
        compare_norms(X_out, self.X.T)



# -*- coding: utf-8 -*-
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

# Plugin to read the mountainsmap surface format (sur)
#Current state can bring support to the surface format if the file is an
#attolight hyperspectral map, but cannot bring write nor support for other
#mountainsmap files (.pro etc.). I need to write some tests, check whether the
#comments can be systematically parsed into metadata and write a support for
#original_metadata or other

import logging
#Dateutil allows to parse date but I don't think it's useful here
#import dateutil.parser

import numpy as np
#Commented for now because I don't know what purpose it serves
#import traits.api as t

from copy import deepcopy
import struct
import sys
import zlib
import os
import warnings

#Maybe later we can implement reading the class with the io utils tools instead
#of re-defining read functions in the class
#import hyperspy.misc.io.utils_readfile as iou

#This module will prove useful when we write the export function
#import hyperspy.misc.io.tools

#DictionaryTreeBrowser class handles the fancy metadata dictionnaries
#from hyperspy.misc.utils import DictionaryTreeBrowser
from hyperspy.exceptions import MountainsMapFileError

_logger = logging.getLogger(__name__)

# Plugin characteristics
# ----------------------
format_name = 'Digital Surf Surface sur'
description = """Read data from the proprietary .sur file format from Digital
Surf. Allows hyperspy to interact with the mountains map software"""
full_support = False #Check with the boys once this is done
# Recognised file extension
file_extensions = ('sur', 'SUR','pro','PRO')
default_extension = 0
# Writing features
writes = False #First we will check with the load
# ----------------------

class DigitalSurfHandler(object):
    """ Class to read Digital Surf MountainsMap files.

    Attributes
    ----------
    filename, signal_dict, _work_dict, _list_sur_file_content, _Object_type,
    _N_data_object, _N_data_channels, _initialized

    Methods
    -------
    parse_file, parse_header, get_image_dictionaries

    Class Variables
    ---------------
    _object_type : dict key: int containing the mountainsmap object types

    """
    #Object types
    _mountains_object_types = {
                                -1: "_ERROR"              ,
                                 0: "_UNKNOWN"            ,
                                 1: "_PROFILE"            ,
                                 2: "_SURFACE"            ,
                                 3: "_BINARYIMAGE"        ,
                                 4: "_PROFILESERIE"       ,
                                 5: "_SURFACESERIE"       ,
                                 6: "_MERIDIANDISC"       ,
                                 7: "_MULTILAYERPROFILE"  ,
                                 8: "_MULTILAYERSURFACE"  ,
                                 9: "_PARALLELDISC"       ,
                                10: "_INTENSITYIMAGE"     ,
                                11: "_INTENSITYSURFACE"   ,
                                12: "_RGBIMAGE"           ,
                                13: "_RGBSURFACE"         ,
                                14: "_FORCECURVE"         ,
                                15: "_SERIEOFFORCECURVE"  ,
                                16: "_RGBINTENSITYSURFACE",
                                20: "_SPECTRUM"           ,
                                21: "_HYPCARD"            ,
                                }

    def __init__(self, filename=None):

        #We do not need to check for file existence here because
        #io module implements it in the load function
        self.filename = filename

        #The signal_dict dictionnary has to be returned by the
        #file_reader function. Apparently original_metadata needs
        #to be set
        self.signal_dict = {'data': np.empty((0,0,0)),
                            'axes': [],
                            'metadata': {},
                            'original_metadata': {}
                            }

        #Dictionary to store, read and write fields in the binary file
        #defined in the MountainsMap SDK. Structure is
        # _work_dict['Field']['value'] : access field value
        # _work_dict['Field']['b_unpack_fn'](f) : unpack value from file f
        # _work_dict['Field']['b_pack_fn'](f,v): pack value v in file f
        self._work_dict = \
            {
            "_01_Signature":
                {
                'value':"DSCOMPRESSED",
                'b_unpack_fn': lambda f: self._get_str(f,12,"DSCOMPRESSED"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,12),
                },
            "_02_Format":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_03_Number_of_Objects":
                {
                'value':1,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_04_Version":
                {
                'value':1,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_05_Object_Type":
                {
                'value':2,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_06_Object_Name":
                {
                'value':"",
                'b_unpack_fn': lambda f: self._get_str(f,30,"DOSONLY"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,30),
                },
            "_07_Operator_Name":
                {
                'value':"",
                'b_unpack_fn': lambda f: self._get_str(f,30,""),
                'b_pack_fn': lambda f,v: self._set_str(f,v,30),
                },
            "_08_P_Size":
                {
                'value':1,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_09_Acquisition_Type":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_10_Range_Type":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_11_Special_Points":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_12_Absolute":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_13_Gauge_Resolution":
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_14_W_Size":
                {
                'value':0,
                'b_unpack_fn': self._get_int32,
                'b_pack_fn': self._set_int32,
                },
            "_15_Size_of_Points":
                {
                'value':16,
                'b_unpack_fn':lambda f: self._get_int16(f,32),
                'b_pack_fn': self._set_int16,
                },
            "_16_Zmin":
                {
                'value':0,
                'b_unpack_fn':self._get_int32,
                'b_pack_fn':self._set_int32,
                },
            "_17_Zmax":
                {
                'value':0,
                'b_unpack_fn':self._get_int32,
                'b_pack_fn': self._set_int32,
                },
            "_18_Number_of_Points":
                {
                'value':0,
                'b_unpack_fn': self._get_int32,
                'b_pack_fn': self._set_int32,
                },
            "_19_Number_of_Lines":
                {
                'value':0,
                'b_unpack_fn':self._get_int32,
                'b_pack_fn':self._set_int32,
                },
            "_20_Total_Nb_of_Pts":
                {
                'value':0,
                'b_unpack_fn': self._get_int32,
                'b_pack_fn': self._set_int32
                },
            "_21_X_Spacing":
                {
                'value':1.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_22_Y_Spacing":
                {
                'value':1.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                 },
            "_23_Z_Spacing":
                {
                'value':1.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_24_Name_of_X_Axis":
                {
                'value':'X',
                'b_unpack_fn': lambda f: self._get_str(f,16,"X"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_25_Name_of_Y_Axis":
                {
                'value':'Y',
                'b_unpack_fn': lambda f: self._get_str(f,16,"Y"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_26_Name_of_Z_Axis":
                {
                'value':'Z',
                'b_unpack_fn': lambda f: self._get_str(f,16,"Z"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_27_X_Step_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,16,"um"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_28_Y_Step_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,16,"um"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_29_Z_Step_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,16,"um"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_30_X_Length_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,16,"um"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_31_Y_Length_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,16,"um"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_32_Z_Length_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,16,"um"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,16),
                },
            "_33_X_Unit_Ratio":
                {
                'value':1.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_34_Y_Unit_Ratio":
                {
                'value':1.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_35_Z_Unit_Ratio":
                {
                'value':1.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_36_Imprint":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_37_Inverted":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_38_Levelled":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_39_Obsolete":
                {
                'value':0,
                'b_unpack_fn': lambda f: self._get_bytes(f,12),
                'b_pack_fn': lambda f,v: self._set_bytes(f,v,12),
                },
            "_40_Seconds":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_41_Minutes":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_42_Hours":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_43_Day":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_44_Month":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_45_Year":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_46_Day_of_week":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_47_Measurement_duration":
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_48_Compressed_data_size":
                {
                'value':0,
                'b_unpack_fn':self._get_uint32,
                'b_pack_fn':self._set_uint32,
                },
            "_49_Obsolete":
                {
                'value':0,
                'b_unpack_fn': lambda f: self._get_bytes(f,6),
                'b_pack_fn': lambda f,v: self._set_bytes(f,v,6),
                },
            "_50_Comment_size":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_51_Private_size":
                {
                'value':0,
                'b_unpack_fn': self._get_int16,
                'b_pack_fn': self._set_int16,
                },
            "_52_Client_zone":
                {
                'value':0,
                'b_unpack_fn': lambda f: self._get_bytes(f,128),
                'b_pack_fn': lambda f,v: self._set_bytes(f,v,128),
                },
            "_53_X_Offset":
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_54_Y_Offset":
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_55_Z_Offset":
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_56_T_Spacing":\
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_57_T_Offset":
                {
                'value':0.0,
                'b_unpack_fn': self._get_float,
                'b_pack_fn': self._set_float,
                },
            "_58_T_Axis_Name":
                {
                'value':'T',
                'b_unpack_fn': lambda f: self._get_str(f,13,"Wavelength"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,13),
                },
            "_59_T_Step_Unit":
                {
                'value':'um',
                'b_unpack_fn': lambda f: self._get_str(f,13,"nm"),
                'b_pack_fn': lambda f,v: self._set_str(f,v,13),
                },
            "_60_Comment":
                {
                'value':0,
                'b_unpack_fn': self._unpack_comment,
                'b_pack_fn': self._pack_comment,
                },
            "_61_Private_zone":
                {
                'value':0,
                'b_unpack_fn': self._unpack_private,
                'b_pack_fn': self._pack_private,
                },
            "_62_points":
                {
                'value':0,
                'b_unpack_fn': self._unpack_data,
                'b_pack_fn': lambda f,v: 0, #Not implemented
                },
            }

        #List of all measurement
        self._list_sur_file_content = []

        #The surface files convention is that when saving multiple data
        #objects at once, they are all packed in the same binary file.
        #Every single object contains a full header with all the sections,
        # but only the first one contains the relevant infos about
        #object type, the number of objects in the file and other.
        #Hence they will be made attributes.
        #Object type
        self._Object_type = "_UNKNOWN"

        #Number of data objects in the file.
        self._N_data_object = 1
        self._N_data_channels = 1

    ### Read methods
    def _read_sur_file(self):
        """Read the binary, possibly compressed, content of the surface
        file. Surface files can be encoded as single or a succession
        of objects. The file is thus read iteratively and from metadata of the
        first file """

        with open(self.filename,'rb') as f:
            #We read the first object
            self._read_single_sur_object(f)
            #We append the first object to the content list
            self._append_work_dict_to_content()
            #Lookup how many objects are stored in the file and save
            self._N_data_object = self._get_work_dict_key_value("_03_Number_of_Objects")
            self._N_data_channels = self._get_work_dict_key_value('_08_P_Size')

            #Determine how many objects we need to read
            if self._N_data_channels>0 and self._N_data_object>0:
                N_objects_to_read = self._N_data_channels*self._N_data_object
            elif self._N_data_channels>0:
                N_objects_to_read = self._N_data_channels
            elif self._N_data_object>0:
                N_objects_to_read = self._N_data_object
            else:
                N_objects_to_read = 1

            #Lookup what object type we are dealing with and save
            self._Object_type = \
                DigitalSurfHandler._mountains_object_types[ \
                    self._get_work_dict_key_value("_05_Object_Type")]

            #if more than 1
            if N_objects_to_read > 1:
                #continue reading until everything is done
                for i in range(1,N_objects_to_read):
                    #We read an object
                    self._read_single_sur_object(f)
                    #We append it to content list
                    self._append_work_dict_to_content()

    def _read_single_sur_object(self,file):
        for key,val in self._work_dict.items():
            self._work_dict[key]['value'] = val['b_unpack_fn'](file)

    def _append_work_dict_to_content(self):
        """Save the values stored in the work dict in the surface file list"""
        datadict = deepcopy( \
            {key:val['value'] for key,val in self._work_dict.items()})
        self._list_sur_file_content.append(datadict)

    def _get_work_dict_key_value(self,key):
        return self._work_dict[key]['value']

    ### Signal dictionary methods
    def _build_sur_dict(self):
        """Create a signal dict with an unpacked object"""

        #If the signal is of the type spectrum or hypercard
        if self._Object_type in ["_HYPCARD",]:
            self._build_hyperspectral_map()
        elif self._Object_type in ["_SPECTRUM"]:
            self._build_spectrum()
        elif self._Object_type in ["_PROFILE"]:
            self._build_general_1D_data()
        elif self._Object_type in ["_PROFILESERIE"]:
            self._build_1D_series()
        elif self._Object_type in ["_SURFACE"]:
            self._build_surface()
        elif self._Object_type in ["_SURFACESERIE"]:
            self._build_surface_series()
        elif self._Object_type in ["_MULTILAYERSURFACE"]:
            self._build_surface_series()
        elif self._Object_type in ["_RGBSURFACE"]:
            self._build_RGB_surface()
        elif self._Object_type in ["_RGBIMAGE"]:
            self._build_RGB_image()
        elif self._Object_type in ["_RGBINTENSITYSURFACE"]:
            self._build_RGB_surface()
        elif self._Object_type in ["_BINARYIMAGE"]:
            self._build_surface()
        else:
            raise MountainsMapFileError(self._Object_type \
                + "is not a supported mountain object.")

        return self.signal_dict

    def _build_Xax(self,unpacked_dict,ind=0,nav=False):
        """Return X axis dictionary from an unpacked dict. index int and navigate
        boolean can be optionally passed. Default 0 and False respectively."""
        Xax = { 'name': unpacked_dict['_24_Name_of_X_Axis'],
                'size': unpacked_dict['_18_Number_of_Points'],
                'index_in_array': ind,
                'scale': unpacked_dict['_21_X_Spacing'],
                'offset': unpacked_dict['_53_X_Offset'],
                'units': unpacked_dict['_27_X_Step_Unit'],
                'navigate':nav,
                }
        return Xax

    def _build_Yax(self,unpacked_dict,ind=1,nav=False):
        """Return X axis dictionary from an unpacked dict. index int and navigate
        boolean can be optionally passed. Default 1 and False respectively."""
        Yax = { 'name': unpacked_dict['_25_Name_of_Y_Axis'],
                'size': unpacked_dict['_19_Number_of_Lines'],
                'index_in_array': ind,
                'scale': unpacked_dict['_22_Y_Spacing'],
                'offset': unpacked_dict['_54_Y_Offset'],
                'units': unpacked_dict['_28_Y_Step_Unit'],
                'navigate':nav,
                }
        return Yax

    def _build_Tax(self,unpacked_dict,size_key,ind=0,nav=True):
        """Return T axis dictionary from an unpacked surface object dict.
        Unlike x and y axes, the size key can be determined from various keys:
        _14_W_Size, _15_Size_of_Points or _03_Number_of_Objects. index int
        and navigate boolean can be optionally passed. Default 0 and
        True respectively."""

        #The T axis is somewhat special because it is only defined on series
        #and is thus only navigation. It is only defined on the first object
        #in a serie.
        #Here it needs to be checked that the T axis scale is not 0 Otherwise
        #it raises hyperspy errors
        scale = unpacked_dict['_56_T_Spacing']
        if scale == 0:
            scale =1

        Tax = { 'name': unpacked_dict['_58_T_Axis_Name'],
                'size': unpacked_dict[size_key],
                'index_in_array': ind,
                'scale': scale,
                'offset': unpacked_dict['_57_T_Offset'],
                'units': unpacked_dict['_59_T_Step_Unit'],
                'navigate':nav,
                }
        return Tax

    ### Build methods for individual surface objects
    def _build_hyperspectral_map(self,):
        """Build a hyperspectral map. Hyperspectral maps are single-object
        files with datapoints of _14_W_Size length"""

        #Check that the object contained only one object.
        #Probably overkill at this point but better safe than sorry
        if len(self._list_sur_file_content) != 1:
            raise MountainsMapFileError("Input {:s} File is not of Hyperspectral type".format(self._Object_type))

        #We get the dictionary with all the data
        hypdic = self._list_sur_file_content[0]

        #Add all the axes to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Yax(hypdic,ind=0,nav=True))
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=1,nav=True))
        #Wavelength axis in hyperspectral surface files are stored as T Axis
        #with length set as _14_W_Size
        self.signal_dict['axes'].append(\
            self._build_Tax(hypdic,'_14_W_Size',ind=2,nav=False))

        #We reshape the data in the correct format
        self.signal_dict['data'] = hypdic['_62_points'].reshape(\
            hypdic['_19_Number_of_Lines'],
            hypdic['_18_Number_of_Points'],
            hypdic['_14_W_Size'],
            )

        self.signal_dict['metadata'] = self._build_metadata(hypdic)

        self.signal_dict['original_metadata'] = self._build_original_metadata()

    def _build_general_1D_data(self,):
        """Build general 1D Data objects. Currently work with spectra"""

        #Check that the object contained only one object.
        #Probably overkill at this point but better safe than sorry
        if len(self._list_sur_file_content) != 1:
            raise MountainsMapFileError("Corrupt file")

        #We get the dictionary with all the data
        hypdic = self._list_sur_file_content[0]

        #Add the axe to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=0,nav=False))

        #We reshape the data in the correct format
        self.signal_dict['data'] = hypdic['_62_points']

        #Build the metadata
        self._set_metadata_and_original_metadata(hypdic)

    def _build_spectrum(self,):
        """Build spectra objects. Spectra and 1D series of spectra are
        saved in the same object."""

        #We get the dictionary with all the data
        hypdic = self._list_sur_file_content[0]

        #Add the signal axis_src to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=1,nav=False))

        #If there is more than 1 spectrum also add the navigation axis
        if hypdic['_19_Number_of_Lines'] != 1:
            self.signal_dict['axes'].append(\
                self._build_Yax(hypdic,ind=0,nav=True))

        #We reshape the data in the correct format.
        #Edit: the data is now squeezed for unneeded dimensions
        self.signal_dict['data'] = np.squeeze(hypdic['_62_points'].reshape(\
            hypdic['_19_Number_of_Lines'],
            hypdic['_18_Number_of_Points'],
            ))

        self._set_metadata_and_original_metadata(hypdic)

    def _build_1D_series(self,):
        """Build a series of 1D objects. The T axis is navigation and set from
        the first object"""

        #First object dictionary
        hypdic = self._list_sur_file_content[0]

        #Metadata are set from first dictionary
        self._set_metadata_and_original_metadata(hypdic)

        #Add the series-axis to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Tax(hypdic,'_03_Number_of_Objects',ind=0,nav=True))

        #All objects must share the same signal axis
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=1,nav=False))

        #We put all the data together
        data = []
        for obj in self._list_sur_file_content:
            data.append(obj['_62_points'])

        self.signal_dict['data'] = np.stack(data)

    def _build_surface(self,):
        """Build a surface"""

        #Check that the object contained only one object.
        #Probably overkill at this point but better safe than sorry
        if len(self._list_sur_file_content) != 1:
            raise MountainsMapFileError("CORRUPT {:s} FILE".format(self._Object_type))

        #We get the dictionary with all the data
        hypdic = self._list_sur_file_content[0]

        #Add all the axes to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Yax(hypdic,ind=0,nav=False))
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=1,nav=False))



        #We reshape the data in the correct format
        shape = (hypdic['_19_Number_of_Lines'],hypdic['_18_Number_of_Points'])
        self.signal_dict['data'] = hypdic['_62_points'].reshape(shape)

        self._set_metadata_and_original_metadata(hypdic)

    def _build_surface_series(self,):
        """Build a series of surfaces. The T axis is navigation and set from
        the first object"""

        #First object dictionary
        hypdic = self._list_sur_file_content[0]

        #Metadata are set from first dictionary
        self._set_metadata_and_original_metadata(hypdic)

        #Add the series-axis to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Tax(hypdic,'_03_Number_of_Objects',ind=0,nav=True))

        #All objects must share the same signal axes
        self.signal_dict['axes'].append(\
                self._build_Yax(hypdic,ind=1,nav=False))
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=2,nav=False))

        #shape of the surfaces in the series
        shape = (hypdic['_19_Number_of_Lines'],hypdic['_18_Number_of_Points'])
        #We put all the data together
        data = []
        for obj in self._list_sur_file_content:
            data.append(obj['_62_points'].reshape(shape))

        self.signal_dict['data'] = np.stack(data)

    def _build_RGB_surface(self,):
        """Build a series of surfaces. The T axis is navigation and set from
        P Size"""

        #First object dictionary
        hypdic = self._list_sur_file_content[0]

        #Metadata are set from first dictionary
        self._set_metadata_and_original_metadata(hypdic)

        #Add the series-axis to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Tax(hypdic,'_08_P_Size',ind=0,nav=True))

        #All objects must share the same signal axes
        self.signal_dict['axes'].append(\
                self._build_Yax(hypdic,ind=1,nav=False))
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=2,nav=False))

        #shape of the surfaces in the series
        shape = (hypdic['_19_Number_of_Lines'],hypdic['_18_Number_of_Points'])
        #We put all the data together
        data = []
        for obj in self._list_sur_file_content:
            data.append(obj['_62_points'].reshape(shape))

        #Pushing data into the dictionary
        self.signal_dict['data'] = np.stack(data)

    def _build_RGB_image(self,):
        """Build an RGB image. The T axis is navigation and set from
        P Size"""

        #First object dictionary
        hypdic = self._list_sur_file_content[0]

        #Metadata are set from first dictionary
        self._set_metadata_and_original_metadata(hypdic)

        #Add the series-axis to the signal dict
        self.signal_dict['axes'].append(\
            self._build_Tax(hypdic,'_08_P_Size',ind=0,nav=True))

        #All objects must share the same signal axes
        self.signal_dict['axes'].append(\
                self._build_Yax(hypdic,ind=1,nav=False))
        self.signal_dict['axes'].append(\
            self._build_Xax(hypdic,ind=2,nav=False))

        #shape of the surfaces in the series
        shape = (hypdic['_19_Number_of_Lines'],hypdic['_18_Number_of_Points'])
        #We put all the data together
        data = []
        for obj in self._list_sur_file_content:
            data.append(obj['_62_points'].reshape(shape))

        #Pushing data into the dictionary
        self.signal_dict['data'] = np.stack(data)

        self.signal_dict.update({'post_process':[self.post_process_RGB]})

    ### Metadata utility methods
    def _build_metadata(self,unpacked_dict):
        """Return a minimalistic metadata dictionary according to hyperspy
        format. Accept a dictionary as an input because dictionary with the
        headers of a mountians object.

        Parameters
        ----------
        unpacked_dict: dictionary from the header of a surface file

        Returns
        -------
        metadict: dictionnary in the hyperspy metadata format

        """

        #Formatting for complicated strings. We add parentheses to units
        qty_unit = unpacked_dict['_29_Z_Step_Unit']
        #We strip unit from any character that might pre-format it
        qty_unit = qty_unit.strip(' \t\n()[]')
        #If unit string is still truthy after strip we add parentheses
        if qty_unit:
            qty_unit = "({:s})".format(qty_unit)


        quantity_str = " ".join([
            unpacked_dict['_26_Name_of_Z_Axis'],qty_unit]).strip()

        #Date and time are set in metadata only if all values are not set to 0

        date = [unpacked_dict['_45_Year'],
                unpacked_dict['_44_Month'],
                unpacked_dict['_43_Day']]
        if not all(v == 0 for v in date):
            date_str =  "{:4d}-{:2d}-{:2d}".format(date[0],date[1],date[2])
        else:
            date_str = ""

        time = [unpacked_dict['_42_Hours'],
                unpacked_dict['_41_Minutes'],
                unpacked_dict['_40_Seconds']]

        if not all(v == 0 for v in time):
            time_str = "{:d}:{:d}:{:d}".format(time[0],time[1],time[2])
        else:
            time_str = ""

        #Metadata dictionary initialization
        metadict = {
            "General":{
                "authors": unpacked_dict['_07_Operator_Name'],
                "date":date_str,
                "original_filename": os.path.split(self.filename)[1],
                "time": time_str,
                },
            "Signal": {
                "binned": False,
                "quantity": quantity_str,
                "signal_type": "",
                },
            }

        return metadict

    def _build_original_metadata(self,):
        """Builds a metadata dictionnary from the header"""
        original_metadata_dict = {}

        Ntot = (self._N_data_object+1)*(self._N_data_channels+1)

        #Iteration over Number of data objects
        for i in range(self._N_data_object):
            #Iteration over the Number of Data channels
            for j in range(self._N_data_channels):
                #Creating a dictionary key for each object
                k = (i+1)*(j+1)
                key = "Object_{:d}_Channel_{:d}".format(i,j)
                original_metadata_dict.update({key:{}})

                #We load one full object header
                a = self._list_sur_file_content[k-1]

                #Save it as original metadata dictionary
                headerdict = {"H"+l.lstrip('_'):a[l] for l in a if l not in \
                    ("_62_points",'_61_Private_zone')}
                original_metadata_dict[key].update({"Header" : headerdict})

                #The second dictionary might contain custom mountainsmap params
                parsedict = {}

                #Check if it is the case and append it to
                #original metadata if yes

                valid_comment = self._check_comments(a["_60_Comment"],'$','=')
                if valid_comment:
                    parsedict = self._MS_parse(a["_60_Comment"],'$','=')
                    parsedict = {l.lstrip('_'):m for l,m in parsedict.items()}
                    original_metadata_dict[key].update({"Parsed" : parsedict})

        return original_metadata_dict

    def _set_metadata_and_original_metadata(self,unpacked_dict):
        """Run successively _build_metadata and _build_original_metadata
        and set signal dictionary with results"""

        self.signal_dict['metadata'] = self._build_metadata(unpacked_dict)
        self.signal_dict['original_metadata'] = self._build_original_metadata()

    def _check_comments(self,commentsstr,prefix,delimiter):
        """Check if comment string is parsable into metadata dictionary.
        Some specific lines (empty or starting with @@) will be ignored,
        but any non-ignored line must conform to being a title line (beginning
        with the TITLESTART indicator) or being parsable (starting with Prefix
        and containing the key data delimiter). At the end, the comment is
        considered parsable if it contains minimum 1 parsable line and no
        non-ignorable non-parsable non-title line.

        Parameters
        ----------
        commentstr: string containing comments
        prefix: string (or char) character assumed to start each line.
        '$' if a .sur file.
        delimiter: string that delimits the keyword from value. always '='

        Returns
        -------
        valid: boolean
        """

        #Titlestart markers start with Prefix ($) followed by underscore
        TITLESTART = '{:s}_'.format(prefix)

        #We start by assuming that the comment string is valid
        #but contains 0 valid (= parsable) lines
        valid = True
        N_valid_lines = 0

        for line in commentsstr.splitlines():
            #Here we ignore any empty line or line starting with @@
            ignore = False
            if not line.strip() or line.startswith('@@'):
                ignore = True
            #If the line must not be ignored
            if not ignore:
                #If line starts with a titlestart marker we it counts as valid
                if line.startswith(TITLESTART):
                    N_valid_lines += 1
                # if it does not we check that it has the delimiter and
                # starts with prefix
                else:
                    #We check that line contains delimiter and prefix
                    #if it does the count of valid line is increased
                    if delimiter in line and line.startswith(prefix):
                        N_valid_lines += 1
                    #Otherwise the whole comment string is thrown out
                    else:
                        valid = False

        #finally, it total number of valid line is 0 we throw out this comments
        if N_valid_lines ==0:
            valid = False

        #return falsiness of the string.
        return valid

    def _MS_parse(self, strMS, prefix, delimiter):
        """ Parses a string containing metadata information. The string can be
        read from the comment section of a .sur file, or, alternatively, a file
        containing them with a similar formatting.

        Parameters
        ----------
        strMS: string containing metadata
        prefix: string (or char) character assumed to start each line.
        '$' if a .sur file.
        delimiter: string that delimits the keyword from value. always '='

        Returns
        -------
        dictMS: dictionnary in the correct hyperspy metadata format

        """
        #dictMS is created as an empty dictionnary
        dictMS = {}
        #Title lines start with an underscore
        TITLESTART = '{:s}_'.format(prefix)

        for line in strMS.splitlines() :
            #Here we ignore any empty line or line starting with @@
            ignore = False
            if not line.strip() or line.startswith('@@'):
                ignore = True
            #If the line must not be ignored
            if not ignore:
                if line.startswith(TITLESTART):
                    #We strip keys from whitespace at the end and beginning
                    keyMain = line[len(TITLESTART):].strip()
                    dictMS[keyMain] = {}
                elif line.startswith(prefix):
                    key, *liValue = line.split(delimiter)
                    #Key is also stripped from beginning or end whitespace
                    key = key[len(prefix):].strip()
                    strValue = liValue[0] if len(liValue)>0 else ""
                    # remove whitespace at the beginning of value
                    strValue = strValue.strip()
                    liValue = strValue.split(' ')
                    try :
                        if key == "Grating":
                            dictMS[keyMain][key] = liValue[0] # we don't want to eval this one
                        else :
                            dictMS[keyMain][key] = eval(liValue[0])
                    except :
                        dictMS[keyMain][key] = liValue[0]
                    if len(liValue) > 1:
                        dictMS[keyMain][key+'_units'] = liValue[1]
        return dictMS

    ### Post processing
    def post_process_RGB(self,signal):
        signal = signal.transpose()
        max_data = np.nanmax(signal.data)
        if max_data <=256:
            signal.change_dtype('uint8')
            signal.change_dtype('rgb8')
        elif max_data <=65536:
            signal.change_dtype('uint8')
            signal.change_dtype('rgb8')
        else:
            warnings.warn("""RGB-announced data could not be converted to
            uint8 or uint16 datatype""")
            pass

        return signal

    ### pack/unpack binary quantities
    def _get_int16(self,file, default=None, signed=True):
        """Read a 16-bits int with a user-definable default value if
        no file is given"""
        if file is None :
            return default
        b = file.read(2)
        if sys.byteorder == 'big' :
            return struct.unpack('>h', b)[0]
        else :
            return struct.unpack('<h', b)[0]

    def _set_int16(self, file, val):
	       file.write(struct.pack('<h', val))

    def _get_str(self, file, size, default=None, encoding='latin-1'):
        """Read a str of defined size in bytes with a user-definable default
        value if no file is given"""
        if file is None :
            return default
        read_str = file.read(size).decode(encoding)
        return read_str.strip(' \t\n')

    def _set_str(self, file, val, size, encoding='latin-1'):
        """Write a str of defined size in bytes to a file. struct.pack
        will automatically trim the string if it is too long"""
        file.write(struct.pack('<{:d}s'.format(size),
            '{{:<{:d}s}}'.format(size).format(val).encode(encoding)))

    def _get_int32(self,file, default=None):
        """Read a 32-bits int with a user-definable default value if no
        file is given"""
        if file is None :
            return default
        b = file.read(4)
        if sys.byteorder == 'big' :
            return struct.unpack('>i', b)[0]
        else :
            return struct.unpack('<i', b)[0]

    def _set_int32(self, file, val):
        """Write a 32-bits int in a file f """
        file.write(struct.pack('<i', val))

    def _get_float(self,file,default=None):
        """Read a 4-bytes (single precision) float from a binary file f with a
        default value if no file is given"""
        if file is None:
            return default
        return struct.unpack('<f', file.read(4))[0]

    def _set_float(file, val):
        """write a 4-bytes (single precision) float in a file"""
        file.write(struct.pack('<f', val))

    def _get_uint32(self, file, default=None):
    	if file is None :
    		return default
    	b = file.read(4)
    	if sys.byteorder == 'big' :
    		return struct.unpack('>I', b)[0]
    	else :
    		return struct.unpack('<I', b)[0]

    def _set_uint32(self, file, val):
	       file.write(struct.pack('<I', val))

    def _get_bytes(self, file, size, default=None):
        if file is None:
            return default
        else:
            return file.read(size)

    def _set_bytes(self, file, val, size):
    	file.write(struct.pack('<{:d}s'.format(size), val))

    def _unpack_comment(self,file,encoding='latin-1'):
        commentsize = self._get_work_dict_key_value("_50_Comment_size")
        return self._get_str(file,commentsize,encoding)

    def _pack_comment(self,file,val,encoding='latin-1'):
        commentsize = self._get_work_dict_key_value("_50_Comment_size")
        self._set_str(file,val,commentsize)

    def _unpack_private(self,file,encoding='latin-1'):
        privatesize = self._get_work_dict_key_value("_51_Private_size")
        return self._get_str(file,privatesize,encoding)

    def _pack_private(self,file,val,encoding='latin-1'):
        privatesize = self._get_work_dict_key_value("_51_Private_size")
        self._set_str(file,val,commentsize)

    def _unpack_data(self,file,encoding='latin-1'):
        """This needs to be special because it reads until the end of
        file. This causes an error in the series of data"""

        #Size of datapoints in bytes. Always int16 (==2) or 32 (==4)
        Psize = int(self._get_work_dict_key_value('_15_Size_of_Points')/8)
        dtype = np.int16 if Psize == 2 else np.int32

        if self._get_work_dict_key_value('_01_Signature') != 'DSCOMPRESSED' :
            #If the points are not compressed we need to read the exact
            #size occupied by datapoints

            #Datapoints in X and Y dimensions
            Npts_tot = self._get_work_dict_key_value('_20_Total_Nb_of_Pts')
            #Datasize in WL
            Wsize = self._get_work_dict_key_value('_14_W_Size')

            #We need to take into account the fact that Wsize is often
            #set to 0 instead of 1 in non-spectral data to compute the
            #space occupied by data in the file
            readsize = Npts_tot*Psize
            if Wsize != 0:
                readsize*=Wsize
            #if Npts_channel is not 0:
            #    readsize*=Npts_channel

            #Read the exact size of the data
            _points = np.frombuffer(file.read(readsize),dtype=dtype)
            #_points = np.fromstring(file.read(readsize),dtype=dtype)

        else:
            #If the points are compressed do the uncompress magic. There
            #the space occupied by datapoints is self-taken care of.
            #Number of streams
            _directoryCount = self._get_uint32(file)

            #empty lists to store the read sizes
            rawLengthData = []
            zipLengthData = []
            for i in range(_directoryCount):
                #Size of raw and compressed data sizes in each stream
                rawLengthData.append(self._get_uint32(file))
                zipLengthData.append(self._get_uint32(file))

            #We now initialize an empty binary string to store the results
            rawData = b''
            for i in range(_directoryCount):
                #And for each stream we uncompress using zip lib
                #and add it to raw string
                rawData += zlib.decompress(file.read(zipLengthData[i]))

            #Finally numpy converts it to a numeric object
            _points = np.frombuffer(rawData, dtype=dtype)
            #_points = np.fromstring(rawData, dtype=dtype)

        # rescale data
        #We set non measured points to nan according to .sur ways
        nm = []
        if self._get_work_dict_key_value("_11_Special_Points") == 1 :
            # has unmeasured points
            nm = _points == self._get_work_dict_key_value("_16_Zmin")-2

        #We set the point in the numeric scale
        _points = _points.astype(np.float) \
            * self._get_work_dict_key_value("_23_Z_Spacing") \
            * self._get_work_dict_key_value("_35_Z_Unit_Ratio") \
            + self._get_work_dict_key_value("_55_Z_Offset")

        _points[nm] = np.nan
        #Return the points, rescaled
        return _points

    def _pack_data(self,file,val,encoding='latin-1'):
        """This needs to be special because it writes until the end of
        file."""
        datasize = self._get_work_dict_key_value("_62_points")
        self._set_str(file,val,datasize)

def file_reader(filename,**kwds):
    """Read a mountainsmap .sur file and return a dictionnary containing the
    information necessary for creating the data object

    Parameters
    ----------
    filename: name of the .sur file to be read

    Returns
    -------
    signal_dict: dictionnary in the appropriate format. The dictionnary can
    contain several keys including 'data', 'axes', 'metadata', 'original_metadata',
    'post_process', 'mapping', 'attributes'.
    """

    ds = DigitalSurfHandler(filename)

    ds._read_sur_file()

    surdict = ds._build_sur_dict()

    return [surdict,]



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging

import numpy as np
import pytest
from hyperspy import signals
from hyperspy.decorators import lazifyTestClass
from hyperspy.misc.utils import stack


@lazifyTestClass
class TestMismatchedCalibration1D:
    def setup_method(self, method):
        self.s1 = signals.Signal1D([1.0, 2.0, 3.0])
        self.s2 = signals.Signal1D([1.0, 2.0, 3.0])

        self.s1.axes_manager[0].scale = 1.0
        self.s1.axes_manager[0].offset = 0.0
        self.s1.axes_manager[0].units = ""

        self.s2.axes_manager[0].scale = 2.0
        self.s2.axes_manager[0].offset = 0.00001
        self.s2.axes_manager[0].units = "mm"

    @pytest.mark.parametrize("axis", [0, None])
    def test_no_warning(self, axis, caplog):
        with caplog.at_level(logging.WARNING):
            _ = stack([self.s1, self.s1], axis=axis)

        assert "Axis calibration mismatch detected along axis 0" not in caplog.text

    @pytest.mark.parametrize("axis", [0, None])
    def test_has_warning(self, axis, caplog):
        with caplog.at_level(logging.WARNING):
            _ = stack([self.s1, self.s2], axis=axis)

        assert "Axis calibration mismatch detected along axis 0" in caplog.text


class TestMismatchedCalibration2D:
    def setup_method(self, method):
        self.s1 = signals.Signal2D(np.ones((2, 3, 4, 5)))
        self.s2 = signals.Signal2D(np.ones((2, 3, 4, 5)))

        self.s1.axes_manager[1].scale = 1.0
        self.s1.axes_manager[1].offset = 0.0
        self.s1.axes_manager[1].units = ""

        self.s2.axes_manager[1].scale = 2.0
        self.s2.axes_manager[1].offset = 0.00001
        self.s2.axes_manager[1].units = "mm"

    @pytest.mark.parametrize("axis", [0, None])
    def test_no_warning(self, axis, caplog):
        with caplog.at_level(logging.WARNING):
            _ = stack([self.s1, self.s1], axis=axis)

        assert "Axis calibration mismatch detected along axis 1" not in caplog.text

    @pytest.mark.parametrize("axis", [0, None])
    def test_has_warning(self, axis, caplog):
        with caplog.at_level(logging.WARNING):
            _ = stack([self.s1, self.s2], axis=axis)

        assert "Axis calibration mismatch detected along axis 1" in caplog.text



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from hyperspy.api_nogui import *
import logging
_logger = logging.getLogger(__name__)

__doc__ = hyperspy.api_nogui.__doc__

try:
    # Register ipywidgets by importing the module
    import hyperspy_gui_ipywidgets
except ImportError:  # pragma: no cover
    from hyperspy.defaults_parser import preferences
    if preferences.GUIs.warn_if_guis_are_missing:
        _logger.warning(
            "The ipywidgets GUI elements are not available, probably because the "
            "hyperspy_gui_ipywidgets package is not installed.")
try:
    # Register traitui UI elements by importing the module
    import hyperspy_gui_traitsui
except ImportError:  # pragma: no cover
    from hyperspy.defaults_parser import preferences
    if preferences.GUIs.warn_if_guis_are_missing:
        _logger.warning(
            "The traitsui GUI elements are not available, probably because the "
            "hyperspy_gui_traitsui package is not installed.")



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import inspect
from collections.abc import Iterable
from contextlib import contextmanager
from functools import wraps   # Used in exec statement
import re


class Events(object):

    """
    Events container.

    All available events are attributes of this class.

    """

    def __init__(self):
        self._events = {}

    @contextmanager
    def suppress(self):
        """
        Use this function with a 'with' statement to temporarily suppress
        all callbacks of all events in the container. When the 'with' lock
        completes, the old suppression values will be restored.

        Example
        -------
        >>> with obj.events.suppress():
        ...     # Any events triggered by assignments are prevented:
        ...     obj.val_a = a
        ...     obj.val_b = b
        >>> # Trigger one event instead:
        >>> obj.events.values_changed.trigger()

        See also
        --------
        Event.suppress
        Event.suppress_callback
        """

        old = {}

        try:
            for e in self._events.values():
                old[e] = e._suppress
                e._suppress = True
            yield
        finally:
            for e, oldval in old.items():
                e._suppress = oldval

    def _update_doc(self):
        """
        Updates the doc to reflect the events that are contained
        """
        new_doc = self.__class__.__doc__
        new_doc += '\n\tEvents:\n\t-------\n'
        for name, e in self._events.items():
            edoc = inspect.getdoc(e) or ''
            doclines = edoc.splitlines()
            e_short = doclines[0] if len(doclines) > 0 else edoc
            new_doc += '\t%s :\n\t\t%s\n' % (name, e_short)
        new_doc = new_doc.replace('\t', '    ')
        self.__doc__ = new_doc

    def __setattr__(self, name, value):
        """
        Magic to enable having `Event`s as attributes, and keeping them
        separate from other attributes.

        If it's an `Event`, store it in self._events, otherwise set attribute
        in normal way.
        """
        if isinstance(value, Event):
            self._events[name] = value
            self._update_doc()
        else:
            super(Events, self).__setattr__(name, value)

    def __getattr__(self, name):
        """
        Magic to enable having `Event`s as attributes, and keeping them
        separate from other attributes.

        Returns Event attribute `name` (__getattr__ is only called if attribute
        could not be found in the normal way).
        """
        return self._events[name]

    def __delattr__(self, name):
        """
        Magic to enable having `Event`s as attributes, and keeping them
        separate from other attributes.

        Deletes attribute from self._events if present, otherwise delete
        attribute in normal way.
        """
        if name in self._events:
            del self._events[name]
            self._update_doc()
        else:
            super(Events, self).__delattr__(name)

    def __dir__(self):
        """
        Magic to enable having `Event`s as attributes, and keeping them
        separate from other attributes.

        Makes sure tab-completion works in IPython etc.
        """
        d = dir(type(self))
        d.extend(self.__dict__.keys())
        d.extend(self._events.keys())
        return sorted(set(d))

    def __iter__(self):
        """
        Allows iteration of all events in the container
        """
        return self._events.values().__iter__()

    def __repr__(self):
        return "<hyperspy.events.Events: " + repr(self._events) + ">"


class Event(object):

    def __init__(self, doc='', arguments=None):
        """
        Create an Event object.

        Arguments
        ---------
        doc : str
            Optional docstring for the new Event.
        arguments : iterable
            Pass to define the arguments of the trigger() function. Each
            element must either be an argument name, or a tuple containing
            the argument name and the argument's default value.

        Example
        -------
        >>> from hyperspy.events import Event
        >>> Event()
        <hyperspy.events.Event: set()>
        >>> Event(doc="This event has a docstring!").__doc__
        'This event has a docstring!'
        >>> e1 = Event()
        >>> e2 = Event(arguments=('arg1', ('arg2', None)))
        >>> e1.trigger(arg1=12, arg2=43, arg3='str', arg4=4.3)  # Can trigger with whatever
        >>> e2.trigger(arg1=11, arg2=22, arg3=3.4)
        Traceback (most recent call last):
            ...
        TypeError: trigger() got an unexpected keyword argument 'arg3'

        """
        self.__doc__ = doc
        self._arguments = tuple(arguments) if arguments else None
        self._connected_all = set()
        self._connected_some = {}
        self._connected_map = {}
        self._suppress = False
        self._suppressed_callbacks = set()

        if arguments:
            self._trigger_maker(arguments)

    @property
    def arguments(self):
        return self._arguments

    # Regex for confirming valid python identifier
    _re_arg_name = re.compile("[a-zA-Z_][a-zA-Z0-9_]*")

    def _trigger_maker(self, arguments):
        """
        Dynamically creates a function with a signature equal to `arguments`.

        Ensures that trigger can only be called with the correct arguments
        """
        orig_f = self.trigger
        # Validate code for exec!
        defaults = []
        for arg in arguments:
            if isinstance(arg, (tuple, list)):
                defaults.append(arg[1])
                arg = arg[0]
            elif len(defaults) > 0:
                raise SyntaxError(
                    "non-default argument follows default argument")
            m = self._re_arg_name.match(arg)
            if m is None or m.end() != len(arg):
                raise ValueError("Argument name invalid: %s" % arg)
        arguments = [a[0] if isinstance(a, (tuple, list))
                     else a for a in arguments]
        # Create the dynamic code:
        arglist = ', '.join(arguments)
        arg_pass = ', '.join([a + '=' + a for a in arguments])
        wrap_code = u"""
        @wraps(f)
        def trigger(self, %s):
            return f(%s)
        """ % (arglist, arg_pass)
        wrap_code = wrap_code.replace("        ", "")      # Remove indentation
        # Execute dynamic code:
        gl = dict(globals())
        gl.update(locals())
        gl.update({'f': orig_f})    # Make sure it keeps the original!
        exec(wrap_code, gl, locals())
        new_f = locals()['trigger']
        # Replace the trigger function with the new one
        if defaults:
            new_f.__defaults__ = tuple(defaults)
        new_f = new_f.__get__(self, self.__class__)     # Bind method to self
        self.trigger = new_f

    @contextmanager
    def suppress(self):
        """
        Use this function with a 'with' statement to temporarily suppress
        all events in the container. When the 'with' lock completes, the old
        suppression values will be restored.

        Example
        -------
        >>> with obj.events.myevent.suppress():
        ...     # These would normally both trigger myevent:
        ...     obj.val_a = a
        ...     obj.val_b = b

        Trigger manually once:
        >>> obj.events.myevent.trigger()

        See also
        --------
        suppress_callback
        Events.suppress
        """
        old = self._suppress
        self._suppress = True
        try:
            yield
        finally:
            self._suppress = old

    @contextmanager
    def suppress_callback(self, function):
        """
        Use this function with a 'with' statement to temporarily suppress
        a single callback from being called. All other connected callbacks
        will trigger. When the 'with' lock completes, the old suppression value
        will be restored.

        Example
        -------

        >>> with obj.events.myevent.suppress_callback(f):
        ...     # Events will trigger as normal, but `f` will not be called
        ...     obj.val_a = a
        ...     obj.val_b = b
        >>> # Here, `f` will be called as before:
        >>> obj.events.myevent.trigger()

        See also
        --------
        suppress
        Events.suppress
        """
        was_suppressed = function in self._suppressed_callbacks
        if not was_suppressed:
            self._suppressed_callbacks.add(function)
        try:
            yield
        finally:
            if not was_suppressed:
                self._suppressed_callbacks.discard(function)

    @property
    def connected(self):
        """ Connected functions.
        """
        ret = set()
        ret.update(self._connected_all)
        ret.update(self._connected_some.keys())
        ret.update(self._connected_map.keys())
        return ret

    def connect(self, function, kwargs='all'):
        """
        Connects a function to the event.

        Arguments
        ---------
        function : callable
            The function to call when the event triggers.
        kwargs : {tuple or list, dictionary, 'all', 'auto'}, default "all"
            If "all", all the trigger keyword arguments are passed to the
            function. If a list or tuple of strings, only those keyword
            arguments that are in the tuple or list are passed. If empty,
            no keyword argument is passed. If dictionary, the keyword arguments
            of trigger are mapped as indicated in the dictionary. For example,
            {"a" : "b"} maps the trigger argument "a" to the function argument
            "b".

        See also
        --------
        disconnect

        """
        if not callable(function):
            raise TypeError("Only callables can be registered")
        if function in self.connected:
            raise ValueError("Function %s already connected to %s." %
                             (function, self))
        if kwargs == 'auto':
            spec = inspect.signature(function)
            _has_args = False
            _has_kwargs = False
            _normal_params = []
            for name, par in spec.parameters.items():
                if par.kind == par.VAR_POSITIONAL:
                    _has_args = True
                elif par.kind == par.VAR_KEYWORD:
                    _has_kwargs = True
                else:
                    _normal_params.append(name)
            if _has_args and not _has_kwargs:
                raise NotImplementedError("Connecting to variable argument "
                                          "functions is not supported in auto "
                                          "connection mode.")
            elif _has_kwargs:
                kwargs = 'all'
            else:
                kwargs = _normal_params
        if kwargs == "all":
            self._connected_all.add(function)
        elif isinstance(kwargs, dict):
            self._connected_map[function] = kwargs
        elif isinstance(kwargs, (tuple, list)):
            self._connected_some[function] = tuple(kwargs)
        else:
            raise ValueError("Invalid value passed to kwargs.")

    def disconnect(self, function):
        """
        Disconnects a function from the event. The passed function will be
        disconnected irregardless of which 'nargs' argument was passed to
        connect().

        If you only need to temporarily prevent a function from being called,
        single callback suppression is supported by the `suppress_callback`
        context manager.

        Parameters
        ----------
        function: function
        return_connection_kwargs: bool, default False
            If True, returns the kwargs that would reconnect the function as
            it was.

        See also
        --------
        connect
        suppress_callback
        """
        if function in self._connected_all:
            self._connected_all.remove(function)
        elif function in self._connected_some:
            self._connected_some.pop(function)
        elif function in self._connected_map:
            self._connected_map.pop(function)
        else:
            raise ValueError("The %s function is not connected to %s." %
                             (function, self))

    def trigger(self, **kwargs):
        """
        Triggers the event. If the event is suppressed, this does nothing.
        Otherwise it calls all the connected functions with the arguments as
        specified when connected.

        See also
        --------
        suppress
        suppress_callback
        Events.suppress
        """
        if self._suppress:
            return
        # Work on copies of collections of connected functions.
        # Take copies initially, to ensure that all functions connected when
        # event triggered are called.
        connected_all = self._connected_all.difference(
            self._suppressed_callbacks)
        connected_some = list(self._connected_some.items())
        connected_map = list(self._connected_map.items())

        # Loop over all collections
        for function in connected_all:
            function(**kwargs)
        for function, kwsl in connected_some:
            if function not in self._suppressed_callbacks:
                function(**{kw: kwargs.get(kw, None) for kw in kwsl})
        for function, kwsd in connected_map:
            if function not in self._suppressed_callbacks:
                function(**{kwf: kwargs[kwt] for kwt, kwf in kwsd.items()})

    def __deepcopy__(self, memo):
        dc = type(self)()
        memo[id(self)] = dc
        return dc

    def __str__(self):
        if self.__doc__:
            edoc = inspect.getdoc(self) or ''
            doclines = edoc.splitlines()
            e_short = doclines[0] if len(doclines) > 0 else edoc
            text = ("<hyperspy.events.Event: " + e_short + ": " +
                    str(self.connected) + ">")
        else:
            text = self.__repr__()
        return text

    def __repr__(self):
        return "<hyperspy.events.Event: " + repr(self.connected) + ">"


class EventSuppressor(object):

    """
    Object to enforce a variety of suppression types simultaneously

    Targets to be suppressed can be added by the function `add()`, or given
    in the constructor. Valid targets are:

    * `Event`: The entire Event will be suppressed
    * `Events`: All events in th container will be suppressed
    * (Event, callback): The callback will be suppressed in Event
    * (Events, callback): The callback will be suppressed in each event in
      Events where it is connected.
    * Any iterable collection of the above target types

    Example
    -------
    >>> es = EventSuppressor((event1, callback1), (event1, callback2))
    >>> es.add(event2, callback2)
    >>> es.add(event3)
    >>> es.add(events_container1)
    >>> es.add(events_container2, callback1)
    >>> es.add(event4, (events_container3, callback2))
    >>>
    >>> with es.suppress():
    ...     do_something()
    """

    def __init__(self, *to_suppress):
        self._cms = []
        if len(to_suppress) > 0:
            self.add(*to_suppress)

    def _add_single(self, target):
        # Identify and initializes the CM, but doesn't enter it
        if self._is_tuple_target(target):
            if isinstance(target[0], Event):
                cm = target[0].suppress_callback(target[1])
                self._cms.append(cm)
            else:
                # Don't check for function presence in event now:
                # suppress_callback does this when entering
                for e in target[0]:
                    self._cms.append(e.suppress_callback(target[1]))
        else:
            cm = target.suppress()
            self._cms.append(cm)

    def _is_tuple_target(self, candidate):
        v = (isinstance(candidate, Iterable) and
             not isinstance(candidate, Events) and
             len(candidate) == 2 and
             isinstance(candidate[0], (Event, Events)) and
             callable(candidate[1]))
        return v

    def _is_target(self, candidate):
        v = (isinstance(candidate, (Event, Events)) or
             self._is_tuple_target(candidate))
        return v

    def add(self, *to_suppress):
        """
        Add one or more targets to be suppressed

        Valid targets are:
         - `Event`: The entire Event will be suppressed
         - `Events`: All events in the container will be suppressed
         - (Event, callback): The callback will be suppressed in Event
         - (Events, callback): The callback will be suppressed in each event
           in Events where it is connected.
         - Any iterable collection of the above target types
        """
        # Remove useless layers of iterables:
        while (isinstance(to_suppress, Iterable) and
               not isinstance(to_suppress, Events) and len(to_suppress) == 1):
            to_suppress = to_suppress[0]
        # If single target passed, add directly:
        if self._is_target(to_suppress):
            self._add_single(to_suppress)
        elif isinstance(to_suppress, Iterable):
            if len(to_suppress) == 0:
                raise ValueError("No viable suppression targets added!")
            for t in to_suppress:
                if self._is_target(t):
                    self._add_single(t)
        else:
            raise ValueError("No viable suppression targets added!")

    @contextmanager
    def suppress(self):
        """
        Use this function with a 'with' statement to temporarily suppress
        all events added. When the 'with' lock completes, the old suppression
        values will be restored.

        See also
        --------
        Events.suppress
        Event.suppress
        Event.suppress_callback
        """
        # We don't suppress any exceptions, so we can use simple CM management:
        cms = []
        try:
            for cm in self._cms:
                cm.__enter__()
                cms.append(cm)                  # Only add entered CMs to list
            yield
        finally:
            # Completed succefully or exception occured, unwind all
            for cm in reversed(cms):
                # We don't use exception info, so simply pass blanks
                cm.__exit__(None, None, None)



# -*- coding: utf-8 -*-
# Copyright 2007-2016 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import itertools

import numpy as np
import pytest

from hyperspy.components1d import SplitVoigt
from hyperspy.signals import Signal1D
from hyperspy.utils import stack


TRUE_FALSE_2_TUPLE = [p for p in itertools.product((True, False), repeat=2)]


def test_function():
    g = SplitVoigt()
    g.A.value = 2.5
    g.centre.value = 0
    g.sigma1.value = 2
    g.sigma2.value = 2
    np.testing.assert_allclose(g.function(0), 0.49867785, rtol=1e-3)
    np.testing.assert_allclose(g.function(6), 0.00553981, rtol=1e-3)


@pytest.mark.parametrize(("lazy"), (True, False))
@pytest.mark.parametrize(("only_current", "binned"), TRUE_FALSE_2_TUPLE)
def test_estimate_parameters_binned(only_current, binned, lazy):
    s = Signal1D(np.empty((100,)))
    s.metadata.Signal.binned = binned
    axis = s.axes_manager.signal_axes[0]
    axis.scale = 1
    axis.offset = -20
    g1 = SplitVoigt(A=20001.0, centre=10.0, sigma1=3.0, sigma2=3.0)
    s.data = g1.function(axis.axis)
    if lazy:
        s = s.as_lazy()
    g2 = SplitVoigt()
    factor = axis.scale if binned else 1
    assert g2.estimate_parameters(s, axis.low_value, axis.high_value,
                                  only_current=only_current)
    assert g2.binned == binned
    np.testing.assert_allclose(g1.A.value, g2.A.value * factor, rtol=0.2)
    assert abs(g2.centre.value - g1.centre.value) <= 0.1
    assert abs(g2.sigma1.value - g1.sigma1.value) <= 0.1
    assert abs(g2.sigma2.value - g1.sigma2.value) <= 0.1


@pytest.mark.parametrize(("lazy"), (True, False))
def test_function_nd(lazy):
    s = Signal1D(np.empty((200,)))
    axis = s.axes_manager.signal_axes[0]
    axis.scale = .05
    axis.offset = -5
    A, sigma1, sigma2, fraction, centre = 5, 0.3, 0.75, 0.5, 1
    g1 = SplitVoigt(A=A, sigma1=sigma1, sigma2=sigma2, fraction=fraction,
                    centre=centre)
    s.data = g1.function(axis.axis)
    s2 = stack([s] * 2)
    if lazy:
        s2 = s2.as_lazy()
    g2 = SplitVoigt()
    assert g2.estimate_parameters(s2, axis.low_value, axis.high_value, False)

    g2.A.map['values'] = [A] * 2
    g2.sigma1.map['values'] = [sigma1] * 2
    g2.sigma2.map['values'] = [sigma2] * 2
    g2.fraction.map['values'] = [fraction] * 2
    g2.centre.map['values'] = [centre] * 2

    np.testing.assert_allclose(g2.function_nd(axis.axis), s2.data)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from hyperspy.api import load
from hyperspy.decorators import lazifyTestClass


@lazifyTestClass
class TestFindPeaks1DOhaver:
    def setup_method(self, method):
        filepath = (
            Path(__file__)
            .resolve()
            .parent.joinpath("data/test_find_peaks1D_ohaver.hdf5")
        )
        self.signal = load(filepath)

    def test_find_peaks1D_ohaver_high_amp_thres(self):
        signal1D = self.signal
        peak_list = signal1D.find_peaks1D_ohaver(amp_thresh=10.0)[0]
        assert len(peak_list) == 0

    def test_find_peaks1D_ohaver_zero_value_bug(self):
        signal1D = self.signal
        peak_list = signal1D.find_peaks1D_ohaver()[0]
        assert len(peak_list) == 48



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.components1d import Gaussian, Lorentzian
from hyperspy.signals import Signal1D
from hyperspy.utils.model_selection import AIC, BIC, AICc


class TestModelSelection:
    def setup_method(self, method):
        s = Signal1D(range(10))
        m1 = s.create_model()
        m2 = s.create_model()
        m1.append(Gaussian())
        m2.append(Lorentzian())
        m1.fit()
        m2.fit()
        self.m1 = m1
        self.m2 = m2

    def test_AIC(self):
        _aic1 = AIC(self.m1)
        _aic2 = AIC(self.m2)
        np.testing.assert_allclose(_aic1, 74.477061729373233)
        np.testing.assert_allclose(_aic2, 72.749265802224159)

    def test_AICc(self):
        _aicc1 = AICc(self.m1)
        _aicc2 = AICc(self.m2)
        np.testing.assert_allclose(_aicc1, 82.477061729373233)
        np.testing.assert_allclose(_aicc2, 80.749265802224159)

    def test_BIC(self):
        _bic1 = BIC(self.m1)
        _bic2 = BIC(self.m2)
        np.testing.assert_allclose(_bic1, 75.687402101349420)
        np.testing.assert_allclose(_bic2, 73.959606174200346)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import math

import numpy as np

from hyperspy.components2d import Gaussian2D

sigma2fwhm = 2 * np.sqrt(2 * np.log(2))


def test_function():
    g = Gaussian2D()
    g.A.value = 14
    g.sigma_x.value = 1.
    g.sigma_y.value = 2.
    g.centre_x.value = -5.
    g.centre_y.value = -5.
    np.testing.assert_allclose(g.function(-5, -5), 1.1140846)
    np.testing.assert_allclose(g.function(-2, -3), 0.007506643)
    assert g._is2D
    assert g._position_x == g.centre_x
    assert g._position_y == g.centre_y


def test_util_fwhm_set():
    g1 = Gaussian2D()
    g1.fwhm_x = 0.33
    g1.fwhm_y = 0.33
    g1.A.value = 1.0
    np.testing.assert_allclose(g1.fwhm_x, g1.sigma_x.value * sigma2fwhm)
    np.testing.assert_allclose(g1.fwhm_y, g1.sigma_y.value * sigma2fwhm)


def test_util_fwhm_get():
    g1 = Gaussian2D(sigma_x=0.33, sigma_y=0.33)
    g1.A.value = 1.0
    np.testing.assert_allclose(g1.fwhm_x, g1.sigma_x.value * sigma2fwhm)
    np.testing.assert_allclose(g1.fwhm_y, g1.sigma_y.value * sigma2fwhm)


def test_util_fwhm_getset():
    g1 = Gaussian2D()
    g1.fwhm_x = 0.33
    g1.fwhm_y = 0.33
    assert g1.fwhm_x == 0.33
    assert g1.fwhm_y == 0.33


def test_properties():
    g = Gaussian2D(add_rotation=True)
    angle = np.radians(20)
    g.rotation_angle.value = angle
    np.testing.assert_allclose(g.rotation_angle_wrapped, angle)

    angle = np.radians(380)
    g.rotation_angle.value = angle
    np.testing.assert_allclose(g.rotation_angle_wrapped, math.fmod(angle, 2 * np.pi))

    g = Gaussian2D(add_rotation=True)
    g.sigma_x.value = 0.5
    g.sigma_y.value = 0.1
    assert g.ellipticity == 5.0
    assert g.rotation_angle.value == 0
    assert g.sigma_major == 0.5
    assert g.sigma_minor == 0.1
    angle = np.radians(20)
    g.rotation_angle.value = angle
    np.testing.assert_allclose(g.rotation_angle_wrapped, angle)
    np.testing.assert_allclose(g.rotation_major_axis, angle)

    g = Gaussian2D(add_rotation=True)
    g.sigma_x.value = 0.1
    g.sigma_y.value = 0.5
    assert g.ellipticity == 5.0
    assert g.rotation_angle.value == 0
    assert g.sigma_major == 0.5
    assert g.sigma_minor == 0.1
    angle = np.radians(20)
    g.rotation_angle.value = angle
    np.testing.assert_allclose(g.rotation_angle_wrapped, angle)
    np.testing.assert_allclose(g.rotation_major_axis, angle - np.pi / 2)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

"""Common docstring snippets for plot.

"""


BASE_PLOT_DOCSTRING_PARAMETERS = \
    """navigator : str, None, or :py:class:`~hyperspy.signal.BaseSignal` (or subclass)
        Allowed string values are ``'auto'``, ``'slider'``, and ``'spectrum'``.

        If ``'auto'``:

                - If `navigation_dimension` > 0, a navigator is
                  provided to explore the data.
                - If `navigation_dimension` is 1 and the signal is an image
                  the navigator is a sum spectrum obtained by integrating
                  over the signal axes (the image).
                - If `navigation_dimension` is 1 and the signal is a spectrum
                  the navigator is an image obtained by stacking all the
                  spectra in the dataset horizontally.
                - If `navigation_dimension` is > 1, the navigator is a sum
                  image obtained by integrating the data over the signal axes.
                - Additionally, if `navigation_dimension` > 2, a window
                  with one slider per axis is raised to navigate the data.
                - For example, if the dataset consists of 3 navigation axes `X`,
                  `Y`, `Z` and one signal axis, `E`, the default navigator will
                  be an image obtained by integrating the data over `E` at the
                  current `Z` index and a window with sliders for the `X`, `Y`,
                  and `Z` axes will be raised. Notice that changing the `Z`-axis
                  index changes the navigator in this case.

            If ``'slider'``:

                - If `navigation dimension` > 0 a window with one slider per
                  axis is raised to navigate the data.

            If ``'spectrum'``:

                - If `navigation_dimension` > 0 the navigator is always a
                  spectrum obtained by integrating the data over all other axes.

            If ``None``, no navigator will be provided.

            Alternatively a :py:class:`~hyperspy.signal.BaseSignal` (or subclass)
            instance can be provided. The `signal_dimension` must be 1 (for a
            spectrum navigator) or 2 (for a image navigator) and
            `navigation_shape` must be 0 (for a static navigator) or
            `navigation_shape` + `signal_shape` must be equal to the
            `navigator_shape` of the current object (for a dynamic navigator).
            If the signal `dtype` is RGB or RGBA this parameter has no effect and
            the value is always set to ``'slider'``.
        axes_manager : None or :py:class:`~hyperspy.axes.AxesManager`
            If None, the signal's `axes_manager` attribute is used.
        plot_markers : bool, default True
            Plot markers added using s.add_marker(marker, permanent=True).
            Note, a large number of markers might lead to very slow plotting.
        navigator_kwds : dict
            Only for image navigator, additional keyword arguments for
            :py:func:`matplotlib.pyplot.imshow`.
        """


BASE_PLOT_DOCSTRING = \
    """Plot the signal at the current coordinates.

        For multidimensional datasets an optional figure,
        the "navigator", with a cursor to navigate that data is
        raised. In any case it is possible to navigate the data using
        the sliders. Currently only signals with signal_dimension equal to
        0, 1 and 2 can be plotted.

        Parameters
        ----------
        """


PLOT1D_DOCSTRING = \
    """norm : str, optional
            The function used to normalize the data prior to plotting.
            Allowable strings are: ``'auto'``, ``'linear'``, ``'log'``.
            (default value is ``'auto'``).
            If ``'auto'``, intensity is plotted on a linear scale except when
            ``power_spectrum=True`` (only for complex signals).
        autoscale : str
            The string must contain any combination of the 'x' and 'v'
            characters. If 'x' or 'v' (for values) are in the string, the
            corresponding horizontal or vertical axis limits are set to their
            maxima and the axis limits will reset when the data or the
            navigation indices are changed. Default is 'v'.
        """


PLOT2D_DOCSTRING = \
    """colorbar : bool, optional
            If true, a colorbar is plotted for non-RGB images.
        autoscale : str
            The string must contain any combination of the 'x', 'y' and 'v'
            characters. If 'x' or 'y' are in the string, the corresponding
            axis limits are set to cover the full range of the data at a given
            position. If 'v' (for values) is in the string, the contrast of the
            image will be set automatically according to `vmin` and `vmax` when
            the data or navigation indices change. Default is 'v'.
        saturated_pixels: scalar
            The percentage of pixels that are left out of the bounds.
            For example, the low and high bounds of a value of 1 are the 0.5%
            and 99.5% percentiles. It must be in the [0, 100] range.
            If None (default value), the value from the preferences is used.
        .. deprecated:: 1.6.0
           `saturated_pixels` will be removed in HyperSpy 2.0.0, it is replaced
            by `vmin`, `vmax` and `autoscale`.
        norm : {"auto", "linear", "power", "log", "symlog" or a subclass of
                :py:class:`matplotlib.colors.Normalise`}
            Set the norm of the image to display. If "auto", a linear scale is
            used except if when `power_spectrum=True` in case of complex data
            type. "symlog" can be used to display negative value on a negative
            scale - read :py:class:`matplotlib.colors.SymLogNorm` and the
            `linthresh` and `linscale` parameter for more details.
        vmin, vmax : {scalar, str}, optional
            `vmin` and `vmax` are used to normalise the displayed data. It can
            be a float or a string. If string, it should be formatted as 'xth',
            where 'x' must be an float in the [0, 100] range. 'x' is used to
            compute the x-th percentile of the data. See
            :py:func:`numpy.percentile` for more information.
        gamma : float
            Parameter used in the power-law normalisation when the parameter
            norm="power". Read :py:class:`matplotlib.colors.PowerNorm` for more
            details. Default value is 1.0.
        linthresh : float
            When used with norm="symlog", define the range within which the
            plot is linear (to avoid having the plot go to infinity around
            zero). Default value is 0.01.
        linscale : float
            This allows the linear range (-linthresh to linthresh) to be
            stretched relative to the logarithmic range. Its value is the
            number of powers of base to use for each half of the linear range.
            See :py:class:`matplotlib.colors.SymLogNorm` for more details.
            Defaulf value is 0.1.
        scalebar : bool, optional
            If True and the units and scale of the x and y axes are the same a
            scale bar is plotted.
        scalebar_color : str, optional
            A valid MPL color string; will be used as the scalebar color.
        axes_ticks : {None, bool}, optional
            If True, plot the axes ticks. If None axes_ticks are only
            plotted when the scale bar is not plotted. If False the axes ticks
            are never plotted.
        axes_off : {bool}
            Default is False.
        no_nans : bool, optional
            If True, set nans to zero for plotting.
        centre_colormap : {"auto", True, False}
            If True the centre of the color scheme is set to zero. This is
            specially useful when using diverging color schemes. If "auto"
            (default), diverging color schemes are automatically centred.
        min_aspect : float
            Set the minimum aspect ratio of the image and the figure. To
            keep the image in the aspect limit the pixels are made
            rectangular.
        """


COMPLEX_DOCSTRING = \
    """power_spectrum : bool, default is False.
            If True, plot the power spectrum instead of the actual signal, if
            False, plot the real and imaginary parts of the complex signal.
        representation : {'cartesian' or 'polar'}
            Determines if the real and imaginary part of the complex data is plotted ('cartesian',
            default), or if the amplitude and phase should be used ('polar').
        same_axes : bool, default True
            If True (default) plot the real and
            imaginary parts (or amplitude and phase) in the same figure if
            the signal is one-dimensional.
        fft_shift : bool, default False
            If True, shift the zero-frequency component.
            See :py:func:`numpy.fft.fftshift` for more details.
        """


PLOT2D_KWARGS_DOCSTRING = \
    """**kwargs
            Only when plotting an image: additional (optional) keyword
            arguments for :py:func:`matplotlib.pyplot.imshow`.
        """



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from unittest import mock

import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy.events import Event


class TestInteractive():

    def setup_method(self, method):
        d = np.linspace(3, 10.5)
        d = np.tile(d, (3, 3, 1))
        # data shape (3, 3, 50)
        s = hs.signals.Signal1D(d)
        self.s = s

    def test_interactive_sum(self):
        s = self.s
        e = Event()
        ss = hs.interactive(s.sum, e, axis=0)
        np.testing.assert_array_equal(ss.data, np.sum(s.data, axis=0))
        s.data += 3.2
        assert not np.allclose(ss.data, np.sum(s.data, axis=0))
        e.trigger()
        np.testing.assert_array_equal(ss.data, np.sum(s.data, axis=0))

    def test_interactive_sum_no_out(self):
        s = self.s

        def sumf(axis):
            return s.sum(axis=axis)
        e = Event()
        ss = hs.interactive(sumf, e, axis=0)
        np.testing.assert_array_equal(ss.data, np.sum(s.data, axis=0))
        s.data += 3.2
        assert not np.allclose(ss.data, np.sum(s.data, axis=0))
        e.trigger()
        np.testing.assert_array_equal(ss.data, np.sum(s.data, axis=0))

    def test_interactive_sum_auto_event(self):
        s = self.s
        ss = hs.interactive(s.sum, axis=0)
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=0))
        s.data += 3.2
        assert not np.allclose(ss.data, np.sum(s.data, axis=0))
        s.events.data_changed.trigger(s)
        np.testing.assert_array_equal(ss.data, np.sum(s.data, axis=0))

    def test_chained_interactive(self):
        s = self.s
        e1, e2 = Event(), Event()
        ss = hs.interactive(s.sum, e1, axis=0)
        sss = hs.interactive(ss.sum, e2, axis=0)
        np.testing.assert_allclose(sss.data, np.sum(s.data, axis=(0, 1)))
        s.data += 3.2
        assert not np.allclose(ss.data, np.sum(s.data, axis=(1)))
        e1.trigger()
        np.testing.assert_allclose(ss.data, np.sum(s.data, axis=(1)))
        assert not np.allclose(sss.data, np.sum(s.data, axis=(0, 1)))
        e2.trigger()
        np.testing.assert_allclose(sss.data, np.sum(s.data, axis=(0, 1)))

    def test_recompute(self):
        s = self.s
        e1 = Event()
        e2 = Event()
        ss = hs.interactive(s.sum, e1, recompute_out_event=e2, axis=0)
        # Check eveything as normal first
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
        # Modify axes and data in-place
        s.crop(1, 1)  # data shape (2, 3, 50)
        # Check that data is no longer comparable
        assert ss.data.shape != np.sum(s.data, axis=1).shape
        # Check that normal event raises an exception due to the invalid shape
        with pytest.raises(ValueError):
            e1.trigger()
        # Check that recompute event fixes issue
        e2.trigger()
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
        # Finally, check that axes are updated as they should
        assert ss.axes_manager.navigation_axes[0].offset == 1

    def test_recompute_auto_recompute(self):
        s = self.s
        ss = hs.interactive(s.sum, axis=0)
        # Check eveything as normal first
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
        # Modify axes and data in-place
        m = mock.Mock()
        s.axes_manager.events.any_axis_changed.connect(m.changed)
        s.crop(1, 1)  # data shape (2, 3, 50)
        assert m.changed.called
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
        # Finally, check that axes are updated as they should
        assert ss.axes_manager.navigation_axes[0].offset == 1

    def test_two_update_events(self):
        s = self.s
        e1 = Event()
        e2 = Event()
        ss = hs.interactive(
            s.sum,
            event=(
                e1,
                e2),
            recompute_out_event=None,
            axis=0)
        s.data[:] = 0
        e1.trigger()
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
        s.data[:] = 1
        e2.trigger()
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))

    def test_two_recompute_events(self):
        s = self.s
        e1 = Event()
        e2 = Event()
        ss = hs.interactive(
            s.sum,
            event=None,
            recompute_out_event=(
                e1,
                e2),
            axis=0)
        s.data[:] = 0
        e1.trigger()
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
        s.data[:] = 1
        e2.trigger()
        np.testing.assert_equal(ss.data, np.sum(s.data, axis=1))
